# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Set of tree views / tree-building widgets, some based on jQuery treeview
plugin.
"""


from cubicweb import _

from warnings import warn

from logilab.mtconverter import xml_escape

from cubicweb.utils import make_uid, json
from cubicweb.predicates import adaptable
from cubicweb.view import EntityView
from cubicweb.web.views import baseviews
from cubicweb.web.views.ajaxcontroller import ajaxfunc

def treecookiename(treeid):
    return str('%s-treestate' % treeid)

def _done_init(done, view, row, col):
    """handle an infinite recursion safety belt"""
    if done is None:
        done = set()
    entity = view.cw_rset.get_entity(row, col)
    if entity.eid in done:
        msg = entity._cw._('loop in %(rel)s relation (%(eid)s)') % {
            'rel': entity.cw_adapt_to('ITree').tree_relation,
            'eid': entity.eid
            }
        return None, msg
    done.add(entity.eid)
    return done, entity


class BaseTreeView(baseviews.ListView):
    """base tree view"""
    __regid__ = 'tree'
    __select__ = adaptable('ITree')
    item_vid = 'treeitem'

    def call(self, done=None, **kwargs):
        if done is None:
            done = set()
        super(BaseTreeView, self).call(done=done, **kwargs)

    def cell_call(self, row, col=0, vid=None, done=None, maxlevel=None, klass=None, **kwargs):
        assert maxlevel is None or maxlevel > 0
        done, entity = _done_init(done, self, row, col)
        if done is None:
            # entity is actually an error message
            self.w(u'<li class="badcontent">%s</li>' % entity)
            return
        self.open_item(entity)
        entity.view(vid or self.item_vid, w=self.w, **kwargs)
        if maxlevel is not None:
            maxlevel -= 1
            if maxlevel == 0:
                self.close_item(entity)
                return
        relatedrset = entity.cw_adapt_to('ITree').children(entities=False)
        self.wview(self.__regid__, relatedrset, 'null', done=done,
                   maxlevel=maxlevel, klass=klass, **kwargs)
        self.close_item(entity)

    def open_item(self, entity):
        self.w(u'<li class="%s">\n' % entity.cw_etype.lower())
    def close_item(self, entity):
        self.w(u'</li>\n')


class TreePathView(EntityView):
    """a recursive path view"""
    __regid__ = 'path'
    __select__ = adaptable('ITree')
    item_vid = 'oneline'
    separator = u'&#160;&gt;&#160;'

    def call(self, **kwargs):
        self.w(u'<div class="pathbar">')
        super(TreePathView, self).call(**kwargs)
        self.w(u'</div>')

    def cell_call(self, row, col=0, vid=None, done=None, **kwargs):
        done, entity = _done_init(done, self, row, col)
        if done is None:
            # entity is actually an error message
            self.w(u'<span class="badcontent">%s</span>' % entity)
            return
        parent = entity.cw_adapt_to('ITree').parent()
        if parent:
            parent.view(self.__regid__, w=self.w, done=done)
            self.w(self.separator)
        entity.view(vid or self.item_vid, w=self.w)


class TreeComboBoxView(TreePathView):
    """display folder in edition's combobox"""
    __regid__ = 'combobox'
    item_vid = 'text'
    separator = u' > '

# XXX rename regid to ajaxtree/foldabletree or something like that (same for
# treeitemview)
class TreeView(EntityView):
    """ajax tree view, click to expand folder"""

    __regid__ = 'treeview'
    itemvid = 'treeitemview'
    subvid = 'oneline'
    cssclass = 'treeview widget'
    title = _('tree view')

    def _init_params(self, subvid, treeid, initial_load, morekwargs):
        form = self._cw.form
        if subvid is None:
            subvid = form.pop('treesubvid', self.subvid)  # consume it
        if treeid is None:
            treeid = form.pop('treeid', None)
            if treeid is None:
                treeid = 'throw_away' + make_uid('uid')
        if 'morekwargs' in self._cw.form:
            ajaxargs = json.loads(form.pop('morekwargs'))
            # got unicode & python keywords must be strings
            morekwargs.update(dict((str(k), v)
                                   for k, v in ajaxargs.items()))
        return subvid, treeid

    def _init_headers(self, treeid):
        self._cw.add_css(('jquery-treeview/jquery.treeview.css', 'cubicweb.treeview.css'))
        self._cw.add_js(('cubicweb.ajax.js', 'cubicweb.widgets.js', 'jquery-treeview/jquery.treeview.js'))
        self._cw.html_headers.add_onload(u"""
jQuery("#tree-%s").treeview({toggle: toggleTree, prerendered: true});""" % treeid)

    def call(self, subvid=None, treeid=None,
             initial_load=True, initial_thru_ajax=None, **morekwargs):
        if initial_thru_ajax is not None:
            msg = '[3.24] initial_thru_ajax argument is deprecated'
            warn(msg, DeprecationWarning, stacklevel=2)

        subvid, treeid = self._init_params(subvid, treeid,
                                           initial_load, morekwargs)
        ulid = ' '
        self._init_headers(treeid)
        ulid = ' id="tree-%s"' % treeid
        self.w(u'<ul%s class="%s">' % (ulid, self.cssclass))
        # XXX force sorting on x.sortvalue() (which return dc_title by default)
        # we need proper ITree & co specification to avoid this.
        # (pb when type ambiguity at the other side of the tree relation,
        # unability to provide generic implementation on eg Folder...)
        for i, entity in enumerate(sorted(self.cw_rset.entities(),
                                          key=lambda x: x.sortvalue())):
            if i+1 < len(self.cw_rset):
                morekwargs['is_last'] = False
            else:
                morekwargs['is_last'] = True
            entity.view(self.itemvid, vid=subvid, parentvid=self.__regid__,
                        treeid=treeid, w=self.w, **morekwargs)
        self.w(u'</ul>')

    def cell_call(self, *args, **allargs):
        """ does not makes much sense until you have to invoke
        somentity.view('treeview') """
        allargs.pop('row')
        allargs.pop('col')
        self.call(*args, **allargs)


class FileTreeView(TreeView):
    """specific version of the treeview to display file trees
    """
    __regid__ = 'filetree'
    cssclass = 'treeview widget filetree'
    title = _('file tree view')

    def call(self, subvid=None, treeid=None, initial_load=True, **kwargs):
        super(FileTreeView, self).call(treeid=treeid, subvid='filetree-oneline',
                                       initial_load=initial_load, **kwargs)

class FileItemInnerView(EntityView):
    """inner view used by the TreeItemView instead of oneline view

    This view adds an enclosing <span> with some specific CSS classes
    around the oneline view. This is needed by the jquery treeview plugin.
    """
    __regid__ = 'filetree-oneline'

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        itree = entity.cw_adapt_to('ITree')
        if itree and not itree.is_leaf():
            self.w(u'<div class="folder">%s</div>\n' % entity.view('oneline'))
        else:
            # XXX define specific CSS classes according to mime types
            self.w(u'<div class="file">%s</div>\n' % entity.view('oneline'))


class DefaultTreeViewItemView(EntityView):
    """default treeitem view for entities which don't adapt to ITree"""
    __regid__ = 'treeitemview'

    def cell_call(self, row, col, vid='oneline', treeid=None, **morekwargs):
        assert treeid is not None
        itemview = self._cw.view(vid, self.cw_rset, row=row, col=col)
        last_class = morekwargs['is_last'] and ' class="last"' or ''
        self.w(u'<li%s>%s</li>' % (last_class, itemview))


class TreeViewItemView(EntityView):
    """specific treeitem view for entities which adapt to ITree

    (each item should be expandable if it's not a tree leaf)
    """
    __regid__ = 'treeitemview'
    __select__ = adaptable('ITree')
    default_branch_state_is_open = False

    def open_state(self, eeid, treeid):
        cookies = self._cw.get_cookie()
        treestate = cookies.get(treecookiename(treeid))
        if treestate:
            return str(eeid) in treestate.value.split(':')
        return self.default_branch_state_is_open

    def cell_call(self, row, col, treeid, vid='oneline', parentvid='treeview',
                  is_last=False, **morekwargs):
        w = self.w
        entity = self.cw_rset.get_entity(row, col)
        itree = entity.cw_adapt_to('ITree')
        liclasses = []
        if self._cw.url(includeparams=False) == entity.absolute_url():
            liclasses.append(u'selected')
        is_open = self.open_state(entity.eid, treeid)
        is_leaf = itree is None or itree.is_leaf()
        if is_leaf:
            if is_last:
                liclasses.append('last')
            w(u'<li class="%s">' % u' '.join(liclasses))
        else:
            rql = itree.children_rql() % {'x': entity.eid}
            url = xml_escape(self._cw.build_url('ajax', rql=rql, vid=parentvid,
                                                pageid=self._cw.pageid,
                                                treeid=treeid,
                                                fname='view',
                                                treesubvid=vid,
                                                morekwargs=json.dumps(morekwargs)))
            divclasses = ['hitarea']
            if is_open:
                liclasses.append('collapsable')
                divclasses.append('collapsable-hitarea')
            else:
                liclasses.append('expandable')
                divclasses.append('expandable-hitarea')
            if is_last:
                if is_open:
                    liclasses.append('lastCollapsable')
                    divclasses.append('lastCollapsable-hitarea')
                else:
                    liclasses.append('lastExpandable')
                    divclasses.append('lastExpandable-hitarea')
            if is_open:
                w(u'<li class="%s">' % u' '.join(liclasses))
            else:
                w(u'<li cubicweb:loadurl="%s" class="%s">' % (url, u' '.join(liclasses)))
            if treeid.startswith('throw_away'):
                divtail = ''
            else:
                divtail = """ onclick="asyncRemoteExec('node_clicked', '%s', '%s')" """ % (
                    treeid, entity.eid)
            w(u'<div class="%s"%s></div>' % (u' '.join(divclasses), divtail))

            # add empty <ul> because jquery's treeview plugin checks for
            # sublists presence
            if not is_open:
                w(u'<ul class="placeholder"><li>place holder</li></ul>')
        # the local node info
        self.wview(vid, self.cw_rset, row=row, col=col, **morekwargs)
        if is_open and not is_leaf: #  => rql is defined
            self.wview(parentvid, itree.children(entities=False), subvid=vid,
                       treeid=treeid, initial_load=False, **morekwargs)
        w(u'</li>')



@ajaxfunc
def node_clicked(self, treeid, nodeeid):
    """add/remove eid in treestate cookie"""
    cookies = self._cw.get_cookie()
    statename = treecookiename(treeid)
    treestate = cookies.get(statename)
    if treestate is None:
        self._cw.set_cookie(statename, nodeeid)
    else:
        marked = set(filter(None, treestate.value.split(':')))
        if nodeeid in marked:
            marked.remove(nodeeid)
        else:
            marked.add(nodeeid)
        self._cw.set_cookie(statename, ':'.join(marked))



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

from yams.buildobjs import (EntityType, RelationDefinition, ComputedRelation,
                            Int, Float)


class Employee(EntityType):
    pass


class employees(RelationDefinition):
    subject = 'Company'
    object = 'Employee'


class associates(RelationDefinition):
    subject = 'Company'
    object = 'Employee'


class works_for(ComputedRelation):
    rule = 'O employees S, NOT EXISTS (O associates S)'


class Company(EntityType):
    score = Float(formula='Any AVG(NN) WHERE X employees E, N concerns E, N note NN')
    score100 = Float(formula='Any AVG(NN) WHERE X employees E, N concerns E, N note100 NN')


class Note(EntityType):
    note = Int()
    note100 = Int(formula='Any N*100 WHERE X note N')


class concerns(RelationDefinition):
    subject = 'Note'
    object = 'Employee'


class whatever(ComputedRelation):
    rule = 'S employees E, O associates E'


class renamed(ComputedRelation):
    rule = 'S employees E, O concerns E'


class perm_changes(ComputedRelation):
    __permissions__ = {'read': ('managers',)}
    rule = 'S employees E, O concerns E'



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""abstract box classes for CubicWeb web client"""


from cubicweb import _

from six import add_metaclass

from logilab.mtconverter import xml_escape
from logilab.common.deprecation import class_deprecated, class_renamed

from cubicweb import Unauthorized, role as get_role
from cubicweb.schema import display_name
from cubicweb.predicates import no_cnx, one_line_rset
from cubicweb.view import View
from cubicweb.web import INTERNAL_FIELD_VALUE, stdmsgs
from cubicweb.web.htmlwidgets import (BoxLink, BoxWidget, SideBoxWidget,
                                      RawBoxItem, BoxSeparator)
from cubicweb.web.action import UnregisteredAction


def sort_by_category(actions, categories_in_order=None):
    """return a list of (category, actions_sorted_by_title)"""
    result = []
    actions_by_cat = {}
    for action in actions:
        actions_by_cat.setdefault(action.category, []).append(
            (action.title, action) )
    for key, values in actions_by_cat.items():
        actions_by_cat[key] = [act for title, act in sorted(values, key=lambda x: x[0])]
    if categories_in_order:
        for cat in categories_in_order:
            if cat in actions_by_cat:
                result.append( (cat, actions_by_cat[cat]) )
    for item in sorted(actions_by_cat.items()):
        result.append(item)
    return result


# old box system, deprecated ###################################################

@add_metaclass(class_deprecated)
class BoxTemplate(View):
    """base template for boxes, usually a (contextual) list of possible
    actions. Various classes attributes may be used to control the box
    rendering.

    You may override one of the formatting callbacks if this is not necessary
    for your custom box.

    Classes inheriting from this class usually only have to override call
    to fetch desired actions, and then to do something like  ::

        box.render(self.w)
    """
    __deprecation_warning__ = '[3.10] *BoxTemplate classes are deprecated, use *CtxComponent instead (%(cls)s)'

    __registry__ = 'ctxcomponents'
    __select__ = ~no_cnx()

    categories_in_order = ()
    cw_property_defs = {
        _('visible'): dict(type='Boolean', default=True,
                           help=_('display the box or not')),
        _('order'):   dict(type='Int', default=99,
                           help=_('display order of the box')),
        # XXX 'incontext' boxes are handled by the default primary view
        _('context'): dict(type='String', default='left',
                           vocabulary=(_('left'), _('incontext'), _('right')),
                           help=_('context where this box should be displayed')),
        }
    context = 'left'

    def sort_actions(self, actions):
        """return a list of (category, actions_sorted_by_title)"""
        return sort_by_category(actions, self.categories_in_order)

    def mk_action(self, title, url, escape=True, **kwargs):
        """factory function to create dummy actions compatible with the
        .format_actions method
        """
        if escape:
            title = xml_escape(title)
        return self.box_action(self._action(title, url, **kwargs))

    def _action(self, title, url, **kwargs):
        return UnregisteredAction(self._cw, title, url, **kwargs)

    # formating callbacks

    def boxitem_link_tooltip(self, action):
        if action.__regid__:
            return u'keyword: %s' % action.__regid__
        return u''

    def box_action(self, action):
        klass = getattr(action, 'html_class', lambda: None)()
        return BoxLink(action.url(), self._cw._(action.title),
                       klass, self.boxitem_link_tooltip(action))


class RQLBoxTemplate(BoxTemplate):
    """abstract box for boxes displaying the content of a rql query not
    related to the current result set.
    """

    # to be defined in concrete classes
    rql = title = None

    def to_display_rql(self):
        assert self.rql is not None, self.__regid__
        return (self.rql,)

    def call(self, **kwargs):
        try:
            rset = self._cw.execute(*self.to_display_rql())
        except Unauthorized:
            # can't access to something in the query, forget this box
            return
        if len(rset) == 0:
            return
        box = BoxWidget(self._cw._(self.title), self.__regid__)
        for i, (teid, tname) in enumerate(rset):
            entity = rset.get_entity(i, 0)
            box.append(self.mk_action(tname, entity.absolute_url()))
        box.render(w=self.w)


class UserRQLBoxTemplate(RQLBoxTemplate):
    """same as rql box template but the rql is build using the eid of the
    request's user
    """

    def to_display_rql(self):
        assert self.rql is not None, self.__regid__
        return (self.rql, {'x': self._cw.user.eid})


class EntityBoxTemplate(BoxTemplate):
    """base class for boxes related to a single entity"""
    __select__ = BoxTemplate.__select__ & one_line_rset()
    context = 'incontext'

    def call(self, row=0, col=0, **kwargs):
        """classes inheriting from EntityBoxTemplate should define cell_call"""
        self.cell_call(row, col, **kwargs)

from cubicweb.web.component import AjaxEditRelationCtxComponent, EditRelationMixIn


class EditRelationBoxTemplate(EditRelationMixIn, EntityBoxTemplate):
    """base class for boxes which let add or remove entities linked
    by a given relation

    subclasses should define at least id, rtype and target
    class attributes.
    """
    rtype = None
    def cell_call(self, row, col, view=None, **kwargs):
        self._cw.add_js('cubicweb.ajax.js')
        entity = self.cw_rset.get_entity(row, col)
        title = display_name(self._cw, self.rtype, get_role(self),
                             context=entity.cw_etype)
        box = SideBoxWidget(title, self.__regid__)
        related = self.related_boxitems(entity)
        unrelated = self.unrelated_boxitems(entity)
        box.extend(related)
        if related and unrelated:
            box.append(BoxSeparator())
        box.extend(unrelated)
        box.render(self.w)

    def box_item(self, entity, etarget, rql, label):
        label = super(EditRelationBoxTemplate, self).box_item(
            entity, etarget, rql, label)
        return RawBoxItem(label, liclass=u'invisible')


AjaxEditRelationBoxTemplate = class_renamed(
    'AjaxEditRelationBoxTemplate', AjaxEditRelationCtxComponent,
    '[3.10] AjaxEditRelationBoxTemplate has been renamed to AjaxEditRelationCtxComponent (%(cls)s)')



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""
The :mod:`cubicweb.web.facet` module contains a set of abstract classes to use
as bases to build your own facets

All facet classes inherits from the :class:`AbstractFacet` class, though you'll
usually find some more handy class that do what you want.

Let's see available classes.

Classes you'll want to use
--------------------------
.. autoclass:: cubicweb.web.facet.RelationFacet
.. autoclass:: cubicweb.web.facet.RelationAttributeFacet
.. autoclass:: cubicweb.web.facet.HasRelationFacet
.. autoclass:: cubicweb.web.facet.AttributeFacet
.. autoclass:: cubicweb.web.facet.RQLPathFacet
.. autoclass:: cubicweb.web.facet.RangeFacet
.. autoclass:: cubicweb.web.facet.DateRangeFacet
.. autoclass:: cubicweb.web.facet.BitFieldFacet
.. autoclass:: cubicweb.web.facet.AbstractRangeRQLPathFacet
.. autoclass:: cubicweb.web.facet.RangeRQLPathFacet
.. autoclass:: cubicweb.web.facet.DateRangeRQLPathFacet

Classes for facets implementor
------------------------------
Unless you didn't find the class that does the job you want above, you may want
to skip those classes...

.. autoclass:: cubicweb.web.facet.AbstractFacet
.. autoclass:: cubicweb.web.facet.VocabularyFacet

.. comment: XXX widgets
"""


from cubicweb import _

from functools import reduce
from warnings import warn
from copy import deepcopy
from datetime import datetime, timedelta

from six import text_type, string_types

from logilab.mtconverter import xml_escape
from logilab.common.graph import has_path
from logilab.common.decorators import cached, cachedproperty
from logilab.common.date import datetime2ticks, ustrftime, ticks2datetime
from logilab.common.deprecation import deprecated
from logilab.common.registry import yes

from rql import nodes, utils

from cubicweb import Unauthorized
from cubicweb.schema import display_name
from cubicweb.uilib import css_em_num_value, domid
from cubicweb.utils import make_uid
from cubicweb.predicates import match_context_prop, partial_relation_possible
from cubicweb.appobject import AppObject
from cubicweb.web import RequestError, htmlwidgets


def rtype_facet_title(facet):
    if facet.cw_rset:
        ptypes = facet.cw_rset.column_types(0)
        if len(ptypes) == 1:
            return display_name(facet._cw, facet.rtype, form=facet.role,
                                context=next(iter(ptypes)))
    return display_name(facet._cw, facet.rtype, form=facet.role)

def get_facet(req, facetid, select, filtered_variable):
    return req.vreg['facets'].object_by_id(facetid, req, select=select,
                                           filtered_variable=filtered_variable)

@deprecated('[3.13] filter_hiddens moved to cubicweb.web.views.facets with '
            'slightly modified prototype')
def filter_hiddens(w, baserql, **kwargs):
    from cubicweb.web.views.facets import filter_hiddens
    return filter_hiddens(w, baserql, wdgs=kwargs.pop('facets'), **kwargs)


## rqlst manipulation functions used by facets ################################

def init_facets(rset, select, mainvar=None):
    """Alters in place the <select> for filtering and returns related data.

    Calls :func:`prepare_select` to prepare the syntaxtree for selection and
    :func:`get_filtered_variable` that selects the variable to be filtered and
    drops several parts of the select tree. See each function docstring for
    details.

    :param rset: ResultSet we init facet for.
    :type rset: :class:`~cubicweb.rset.ResultSet`

    :param select: Select statement to be *altered* to support filtering.
    :type select:   :class:`~rql.stmts.Select` from the ``rset`` parameters.

    :param mainvar: Name of the variable we want to filter with facets.
    :type mainvar:  string

    :rtype: (filtered_variable, baserql) tuple.
    :return filtered_variable:  A rql class:`~rql.node.VariableRef`
                                instance as returned by
                                :func:`get_filtered_variable`.

    :return baserql: A string containing the rql before
                     :func:`prepare_select` but after
                     :func:`get_filtered_variable`.
    """
    rset.req.vreg.rqlhelper.annotate(select)
    filtered_variable = get_filtered_variable(select, mainvar)
    baserql = select.as_string(kwargs=rset.args) # before call to prepare_select
    prepare_select(select, filtered_variable)
    return filtered_variable, baserql

def get_filtered_variable(select, mainvar=None):
    """ Return the variable whose name is `mainvar`
    or the first variable selected in column 0
    """
    if mainvar is None:
        vref = next(select.selection[0].iget_nodes(nodes.VariableRef))
        return vref.variable
    return select.defined_vars[mainvar]

def prepare_select(select, filtered_variable):
    """prepare a syntax tree to generate facet filters

    * remove ORDERBY/GROUPBY clauses
    * cleanup selection (remove everything)
    * undefine unnecessary variables
    * set DISTINCT

    Notice unset of LIMIT/OFFSET us expected to be done by a previous call to
    :func:`get_filtered_variable`.
    """
    # cleanup sort terms / group by
    select.remove_sort_terms()
    select.remove_groups()
    # XXX remove aggregat from having
    # selection: only vocabulary entity
    for term in select.selection[:]:
        select.remove_selected(term)
    # remove unbound variables which only have some type restriction
    for dvar in list(select.defined_vars.values()):
        if not (dvar is filtered_variable or dvar.stinfo['relations']):
            select.undefine_variable(dvar)
    # global tree config: DISTINCT, LIMIT, OFFSET
    select.set_distinct(True)

@deprecated('[3.13] use init_facets instead')
def prepare_facets_rqlst(rqlst, args=None):
    assert len(rqlst.children) == 1, 'FIXME: union not yet supported'
    select = rqlst.children[0]
    filtered_variable = get_filtered_variable(select)
    baserql = select.as_string(args)
    prepare_select(select, filtered_variable)
    return filtered_variable, baserql

def prepare_vocabulary_select(select, filtered_variable, rtype, role,
                              select_target_entity=True):
    """prepare a syntax tree to generate a filter vocabulary rql using the given
    relation:
    * create a variable to filter on this relation
    * add the relation
    * add the new variable to GROUPBY clause if necessary
    * add the new variable to the selection
    """
    newvar = _add_rtype_relation(select, filtered_variable, rtype, role)[0]
    if select_target_entity:
        # if select.groupby: XXX we remove groupby now
        #     select.add_group_var(newvar)
        select.add_selected(newvar)
    # add is restriction if necessary
    if filtered_variable.stinfo['typerel'] is None:
        etypes = frozenset(sol[filtered_variable.name] for sol in select.solutions)
        select.add_type_restriction(filtered_variable, etypes)
    return newvar


def insert_attr_select_relation(select, filtered_variable, rtype, role, attrname,
                                sortfuncname=None, sortasc=True,
                                select_target_entity=True):
    """modify a syntax tree to :
    * link a new variable to `filtered_variable` through `rtype` (where filtered_variable has `role`)
    * retrieve only the newly inserted variable and its `attrname`

    Sorting:
    * on `attrname` ascendant (`sortasc`=True) or descendant (`sortasc`=False)
    * on `sortfuncname`(`attrname`) if `sortfuncname` is specified
    * no sort if `sortasc` is None
    """
    cleanup_select(select, filtered_variable)
    var = prepare_vocabulary_select(select, filtered_variable, rtype, role,
                                   select_target_entity)
    attrvar = select.make_variable()
    select.add_relation(var, attrname, attrvar)
    # if query is grouped, we have to add the attribute variable
    #if select.groupby: XXX may not occur anymore
    #    if not attrvar in select.groupby:
    #        select.add_group_var(attrvar)
    if sortasc is not None:
        _set_orderby(select, attrvar, sortasc, sortfuncname)
    # add attribute variable to selection
    select.add_selected(attrvar)
    return var


def cleanup_select(select, filtered_variable):
    """cleanup tree from unnecessary restrictions:
    * attribute selection
    * optional relations linked to the main variable
    * mandatory relations linked to the main variable
    """
    if select.where is None:
        return
    schema = select.root.schema
    toremove = set()
    vargraph = deepcopy(select.vargraph) # graph representing links between variable
    for rel in select.where.get_nodes(nodes.Relation):
        ovar = _may_be_removed(rel, schema, filtered_variable)
        if ovar is not None:
            toremove.add(ovar)
    removed = set()
    while toremove:
        trvar = toremove.pop()
        trvarname = trvar.name
        # remove paths using this variable from the graph
        linkedvars = vargraph.pop(trvarname)
        for ovarname in linkedvars:
            vargraph[ovarname].remove(trvarname)
        # remove relation using this variable
        for rel in trvar.stinfo['relations']:
            if rel in removed:
                # already removed
                continue
            select.remove_node(rel)
            removed.add(rel)
        rel = trvar.stinfo['typerel']
        if rel is not None and not rel in removed:
            select.remove_node(rel)
            removed.add(rel)
        # cleanup groupby clause
        if select.groupby:
            for vref in select.groupby[:]:
                if vref.name == trvarname:
                    select.remove_group_var(vref)
        # we can also remove all variables which are linked to this variable
        # and have no path to the main variable
        for ovarname in linkedvars:
            if ovarname == filtered_variable.name:
                continue
            if not has_path(vargraph, ovarname, filtered_variable.name):
                toremove.add(select.defined_vars[ovarname])


def _may_be_removed(rel, schema, variable):
    """if the given relation may be removed from the tree, return the variable
    on the other side of `variable`, else return None
    Conditions:
    * the relation is an attribute selection of the main variable
    * the relation is optional relation linked to the main variable
    * the relation is a mandatory relation linked to the main variable
      without any restriction on the other variable
    """
    lhs, rhs = rel.get_variable_parts()
    rschema = schema.rschema(rel.r_type)
    if lhs.variable is variable:
        try:
            ovar = rhs.variable
        except AttributeError:
            # constant restriction
            # XXX: X title LOWER(T) if it makes sense?
            return None
        if rschema.final:
            if len(ovar.stinfo['relations']) == 1 \
                   and not ovar.stinfo.get('having'):
                # attribute selection
                return ovar
            return None
        opt = 'right'
        cardidx = 0
    elif getattr(rhs, 'variable', None) is variable:
        ovar = lhs.variable
        opt = 'left'
        cardidx = 1
    else:
        # not directly linked to the main variable
        return None
    if rel.optional in (opt, 'both'):
        # optional relation
        return ovar
    if all(rdef.cardinality[cardidx] in '1+'
           for rdef in rschema.rdefs.values()):
        # mandatory relation without any restriction on the other variable
        for orel in ovar.stinfo['relations']:
            if rel is orel:
                continue
            if _may_be_removed(orel, schema, ovar) is None:
                return None
        return ovar
    return None

def _make_relation(select, variable, rtype, role):
    newvar = select.make_variable()
    if role == 'object':
        rel = nodes.make_relation(newvar, rtype, (variable,), nodes.VariableRef)
    else:
        rel = nodes.make_relation(variable, rtype, (newvar,), nodes.VariableRef)
    return newvar, rel

def _add_rtype_relation(select, variable, rtype, role):
    """add a relation relying `variable` to entities linked by the `rtype`
    relation (where `variable` has `role`)

    return the inserted variable for linked entities.
    """
    newvar, newrel = _make_relation(select, variable, rtype, role)
    select.add_restriction(newrel)
    return newvar, newrel

def _remove_relation(select, rel, var):
    """remove a constraint relation from the syntax tree"""
    # remove the relation
    select.remove_node(rel)
    # remove relations where the filtered variable appears on the
    # lhs and rhs is a constant restriction
    extra = []
    for vrel in var.stinfo['relations']:
        if vrel is rel:
            continue
        if vrel.children[0].variable is var:
            if not vrel.children[1].get_nodes(nodes.Constant):
                extra.append(vrel)
            select.remove_node(vrel)
    return extra

def _set_orderby(select, newvar, sortasc, sortfuncname):
    if sortfuncname is None:
        select.add_sort_var(newvar, sortasc)
    else:
        vref = nodes.variable_ref(newvar)
        vref.register_reference()
        sortfunc = nodes.Function(sortfuncname)
        sortfunc.append(vref)
        term = nodes.SortTerm(sortfunc, sortasc)
        select.add_sort_term(term)

def _get_var(select, varname, varmap):
    try:
        return varmap[varname]
    except KeyError:
        varmap[varname] = var = select.make_variable()
        return var


_prepare_vocabulary_rqlst = deprecated('[3.13] renamed prepare_vocabulary_select')(
    prepare_vocabulary_select)
_cleanup_rqlst = deprecated('[3.13] renamed to cleanup_select')(cleanup_select)


## base facet classes ##########################################################

class AbstractFacet(AppObject):
    """Abstract base class for all facets. Facets are stored in their own
    'facets' registry. They are similar to contextual components since the use
    the following configurable properties:

    * `visible`, boolean flag telling if a facet should be displayed or not

    * `order`, integer to control facets display order

    * `context`, telling if a facet should be displayed in the table form filter
      (context = 'tablefilter') or in the facet box (context = 'facetbox') or in
      both (context = '')

    The following methods define the facet API:

    .. automethod:: cubicweb.web.facet.AbstractFacet.get_widget
    .. automethod:: cubicweb.web.facet.AbstractFacet.add_rql_restrictions

    Facets will have the following attributes set (beside the standard
    :class:`~cubicweb.appobject.AppObject` ones):

    * `select`, the :class:`rql.stmts.Select` node of the rql syntax tree being
      filtered

    * `filtered_variable`, the variable node in this rql syntax tree that we're
      interested in filtering

    Facets implementors may also be interested in the following properties /
    methods:

    .. autoattribute:: cubicweb.web.facet.AbstractFacet.operator
    .. automethod:: cubicweb.web.facet.AbstractFacet.rqlexec
    """
    __abstract__ = True
    __registry__ = 'facets'
    cw_property_defs = {
        _('visible'): dict(type='Boolean', default=True,
                           help=_('display the facet or not')),
        _('order'):   dict(type='Int', default=99,
                           help=_('display order of the facet')),
        _('context'): dict(type='String', default='',
                           # None <-> both
                           vocabulary=(_('tablefilter'), _('facetbox'), ''),
                           help=_('context where this facet should be displayed, '
                                  'leave empty for both')),
        }
    visible = True
    context = ''
    needs_update = False
    start_unfolded = True
    allow_hide = True
    cw_rset = None # ensure facets have a cw_rset attribute

    def __init__(self, req, select=None, filtered_variable=None,
                 **kwargs):
        super(AbstractFacet, self).__init__(req, **kwargs)
        assert select is not None
        assert filtered_variable
        # take care: facet may be retreived using `object_by_id` from an ajax call
        # or from `select` using the result set to filter
        self.select = select
        self.filtered_variable = filtered_variable

    def __repr__(self):
        return '<%s>' % self.__class__.__name__

    def get_widget(self):
        """Return the widget instance to use to display this facet, or None if
        the facet can't do anything valuable (only one value in the vocabulary
        for instance).
        """
        raise NotImplementedError

    def add_rql_restrictions(self):
        """When some facet criteria has been updated, this method is called to
        add restriction for this facet into the rql syntax tree. It should get
        back its value in form parameters, and modify the syntax tree
        (`self.select`) accordingly.
        """
        raise NotImplementedError

    @property
    def operator(self):
        """Return the operator (AND or OR) to use for this facet when multiple
        values are selected.
        """
        # OR between selected values by default
        return self._cw.form.get(xml_escape(self.__regid__) + '_andor', 'OR')

    def rqlexec(self, rql, args=None):
        """Utility method to execute some rql queries, and simply returning an
        empty list if :exc:`Unauthorized` is raised.
        """
        try:
            return self._cw.execute(rql, args)
        except Unauthorized:
            return []

    @property
    def wdgclass(self):
        raise NotImplementedError

    @property
    @deprecated('[3.13] renamed .select')
    def rqlst(self):
        return self.select


class VocabularyFacet(AbstractFacet):
    """This abstract class extend :class:`AbstractFacet` to use the
    :class:`FacetVocabularyWidget` as widget, suitable for facets that may
    restrict values according to a (usually computed) vocabulary.

    A class which inherits from VocabularyFacet must define at least these methods:

    .. automethod:: cubicweb.web.facet.VocabularyFacet.vocabulary
    .. automethod:: cubicweb.web.facet.VocabularyFacet.possible_values
    """
    needs_update = True
    support_and = False

    @property
    def wdgclass(self):
        return FacetVocabularyWidget

    def get_selected(self):
        return frozenset(int(eid) for eid in self._cw.list_form_param(self.__regid__))

    def get_widget(self):
        """Return the widget instance to use to display this facet.

        This implementation expects a .vocabulary method on the facet and
        return a combobox displaying this vocabulary.
        """
        vocab = self.vocabulary()
        if len(vocab) <= 1:
            return None
        wdg = self.wdgclass(self)
        selected = self.get_selected()
        for label, value in vocab:
            wdg.items.append((value, label, value in selected))
        return wdg

    def vocabulary(self):
        """Return vocabulary for this facet, eg a list of 2-uple (label, value).
        """
        raise NotImplementedError

    def possible_values(self):
        """Return a list of possible values (as string since it's used to
        compare to a form value in javascript) for this facet.
        """
        raise NotImplementedError


class RelationFacet(VocabularyFacet):
    """Base facet to filter some entities according to other entities to which
    they are related. Create concrete facet by inheriting from this class an then
    configuring it by setting class attribute described below.

    The relation is defined by the `rtype` and `role` attributes.

    The `no_relation` boolean flag tells if a special 'no relation' value should be
    added (allowing to filter on entities which *do not* have the relation set).
    Default is computed according the relation's cardinality.

    The values displayed for related entities will be:

    * result of calling their `label_vid` view if specified
    * else their `target_attr` attribute value if specified
    * else their eid (you usually want something nicer...)

    When no `label_vid` is set, you will get translated value if `i18nable` is
    set. By default, `i18nable` will be set according to the schema, but you can
    force its value by setting it has a class attribute.

    You can filter out target entity types by specifying `target_type`.

    By default, vocabulary will be displayed sorted on `target_attr` value in an
    ascending way. You can control sorting with:

    * `sortfunc`: set this to a stored procedure name if you want to sort on the
      result of this function's result instead of direct value

    * `sortasc`: boolean flag to control ascendant/descendant sorting

    To illustrate this facet, let's take for example an *excerpt* of the schema
    of an office location search application:

    .. sourcecode:: python

      class Office(WorkflowableEntityType):
          price = Int(description='euros / m2 / HC / HT')
          surface = Int(description='m2')
          has_address = SubjectRelation('PostalAddress',
                                        cardinality='1?',
                                        composite='subject')
          proposed_by = SubjectRelation('Agency')


    We can simply define a facet to filter offices according to the agency
    proposing it:

    .. sourcecode:: python

      class AgencyFacet(RelationFacet):
          __regid__ = 'agency'
          # this facet should only be selected when visualizing offices
          __select__ = RelationFacet.__select__ & is_instance('Office')
          # this facet is a filter on the 'Agency' entities linked to the office
          # through the 'proposed_by' relation, where the office is the subject
          # of the relation
          rtype = 'has_address'
          # 'subject' is the default but setting it explicitly doesn't hurt...
          role = 'subject'
          # we want to display the agency's name
          target_attr = 'name'
    """
    __select__ = partial_relation_possible() & match_context_prop()
    # class attributes to configure the relation facet
    rtype = None
    role = 'subject'
    target_type = None
    target_attr = 'eid'
    # for subclasses parametrization, should not change if you want a
    # RelationFacet
    target_attr_type = 'Int'
    restr_attr = 'eid'
    restr_attr_type = 'Int'
    comparator = '=' # could be '<', '<=', '>', '>='
    # set this to a stored procedure name if you want to sort on the result of
    # this function's result instead of direct value
    sortfunc = None
    # ascendant/descendant sorting
    sortasc = True
    # if you want to call a view on the entity instead of using `target_attr`
    label_vid = None

    # internal purpose
    _select_target_entity = True

    title = property(rtype_facet_title)
    no_relation_label = _('<no relation>')

    def __repr__(self):
        return '<%s on (%s-%s)>' % (self.__class__.__name__, self.rtype, self.role)

    # facet public API #########################################################

    def vocabulary(self):
        """return vocabulary for this facet, eg a list of 2-uple (label, value)
        """
        select = self.select
        select.save_state()
        if self.rql_sort:
            sort = self.sortasc
        else:
            sort = None # will be sorted on label
        try:
            var = insert_attr_select_relation(
                select, self.filtered_variable, self.rtype, self.role,
                self.target_attr, self.sortfunc, sort,
                self._select_target_entity)
            if self.target_type is not None:
                select.add_type_restriction(var, self.target_type)
            try:
                rset = self.rqlexec(select.as_string(), self.cw_rset.args)
            except Exception:
                self.exception('error while getting vocabulary for %s, rql: %s',
                               self, select.as_string())
                return ()
        finally:
            select.recover()
        # don't call rset_vocabulary on empty result set, it may be an empty
        # *list* (see rqlexec implementation)
        values = rset and self.rset_vocabulary(rset) or []
        if self._include_no_relation():
            values.insert(0, (self._cw._(self.no_relation_label), ''))
        return values

    def possible_values(self):
        """return a list of possible values (as string since it's used to
        compare to a form value in javascript) for this facet
        """
        select = self.select
        select.save_state()
        try:
            cleanup_select(select, self.filtered_variable)
            if self._select_target_entity:
                prepare_vocabulary_select(select, self.filtered_variable, self.rtype,
                                         self.role, select_target_entity=True)
            else:
                insert_attr_select_relation(
                    select, self.filtered_variable, self.rtype, self.role,
                    self.target_attr, select_target_entity=False)
            values = [text_type(x) for x, in self.rqlexec(select.as_string())]
        except Exception:
            self.exception('while computing values for %s', self)
            return []
        finally:
            select.recover()
        if self._include_no_relation():
            values.append('')
        return values

    def add_rql_restrictions(self):
        """add restriction for this facet into the rql syntax tree"""
        value = self._cw.form.get(self.__regid__)
        if value is None:
            return
        filtered_variable = self.filtered_variable
        restrvar, rel = _add_rtype_relation(self.select, filtered_variable,
                                            self.rtype, self.role)
        self.value_restriction(restrvar, rel, value)

    # internal control API #####################################################

    @property
    def i18nable(self):
        """should label be internationalized"""
        if self.target_type:
            eschema = self._cw.vreg.schema.eschema(self.target_type)
        elif self.role == 'subject':
            eschema = self._cw.vreg.schema.rschema(self.rtype).objects()[0]
        else:
            eschema = self._cw.vreg.schema.rschema(self.rtype).subjects()[0]
        return getattr(eschema.rdef(self.target_attr), 'internationalizable', False)

    @property
    def no_relation(self):
        return (not self._cw.vreg.schema.rschema(self.rtype).final
                and self._search_card('?*'))

    @property
    def rql_sort(self):
        """return true if we can handle sorting in the rql query. E.g.  if
        sortfunc is set or if we have not to transform the returned value (eg no
        label_vid and not i18nable)
        """
        return self.sortfunc is not None or (self.label_vid is None
                                             and not self.i18nable)

    def rset_vocabulary(self, rset):
        if self.i18nable:
            tr = self._cw._
        else:
            tr = text_type
        if self.rql_sort:
            values = [(tr(label), eid) for eid, label in rset]
        else:
            if self.label_vid is None:
                values = [(tr(label), eid) for eid, label in rset]
            else:
                values = [(entity.view(self.label_vid), entity.eid)
                          for entity in rset.entities()]
            values = sorted(values)
            if not self.sortasc:
                values = list(reversed(values))
        return values

    @property
    def support_and(self):
        return self._search_card('+*')

    # internal utilities #######################################################

    @cached
    def _support_and_compat(self):
        support = self.support_and
        if callable(support):
            warn('[3.13] %s.support_and is now a property' % self.__class__,
                 DeprecationWarning)
            support = support()
        return support

    def value_restriction(self, restrvar, rel, value):
        # XXX handle rel is None case in RQLPathFacet?
        if self.restr_attr != 'eid':
            self.select.set_distinct(True)
        if isinstance(value, string_types):
            # only one value selected
            if value:
                self.select.add_constant_restriction(
                    restrvar, self.restr_attr, value,
                    self.restr_attr_type)
            else:
                rel.parent.replace(rel, nodes.Not(rel))
        elif self.operator == 'OR':
            # set_distinct only if rtype cardinality is > 1
            if self._support_and_compat():
                self.select.set_distinct(True)
            # multiple ORed values: using IN is fine
            if '' in value:
                value.remove('')
                self._add_not_rel_restr(rel)
            self._and_restriction(rel, restrvar, value)
        else:
            # multiple values with AND operator. We've to generate a query like
            # "X relation A, A eid 1, X relation B, B eid 1", hence the new
            # relations at each iteration in the while loop below 
            if '' in value:
                raise RequestError("this doesn't make sense")
            self._and_restriction(rel, restrvar, value.pop())
            while value:
                restrvar, rtrel = _make_relation(self.select, self.filtered_variable,
                                                 self.rtype, self.role)
                if rel is None:
                    self.select.add_restriction(rtrel)
                else:
                    rel.parent.replace(rel, nodes.And(rel, rtrel))
                self._and_restriction(rel, restrvar, value.pop())

    def _and_restriction(self, rel, restrvar, value):
        if rel is None:
            self.select.add_constant_restriction(restrvar, self.restr_attr,
                                                 value, self.restr_attr_type)
        else:
            rrel = nodes.make_constant_restriction(restrvar, self.restr_attr,
                                                   value, self.restr_attr_type)
            rel.parent.replace(rel, nodes.And(rel, rrel))


    @cached
    def _search_card(self, cards):
        for rdef in self._iter_rdefs():
            if rdef.role_cardinality(self.role) in cards:
                return True
        return False

    def _iter_rdefs(self):
        rschema = self._cw.vreg.schema.rschema(self.rtype)
        # XXX when called via ajax, no rset to compute possible types
        possibletypes = self.cw_rset and self.cw_rset.column_types(0)
        for rdef in rschema.rdefs.values():
            if possibletypes is not None:
                if self.role == 'subject':
                    if rdef.subject not in possibletypes:
                        continue
                elif rdef.object not in possibletypes:
                    continue
            if self.target_type is not None:
                if self.role == 'subject':
                    if rdef.object != self.target_type:
                        continue
                elif rdef.subject != self.target_type:
                    continue
            yield rdef

    def _include_no_relation(self):
        if not self.no_relation:
            return False
        if self._cw.vreg.schema.rschema(self.rtype).final:
            return False
        if self.role == 'object':
            subj = next(utils.rqlvar_maker(defined=self.select.defined_vars,
                                      aliases=self.select.aliases))
            obj = self.filtered_variable.name
        else:
            subj = self.filtered_variable.name
            obj = next(utils.rqlvar_maker(defined=self.select.defined_vars,
                                     aliases=self.select.aliases))
        restrictions = []
        if self.select.where:
            restrictions.append(self.select.where.as_string())
        if self.select.with_:
            restrictions.append('WITH ' + ','.join(
                term.as_string() for term in self.select.with_))
        if restrictions:
            restrictions = ',' + ','.join(restrictions)
        else:
            restrictions = ''
        rql = 'Any %s LIMIT 1 WHERE NOT %s %s %s%s' % (
            self.filtered_variable.name, subj, self.rtype, obj, restrictions)
        try:
            return bool(self.rqlexec(rql, self.cw_rset and self.cw_rset.args))
        except Exception:
            # catch exception on executing rql, work-around #1356884 until a
            # proper fix
            self.exception('cant handle rql generated by %s', self)
            return False

    def _add_not_rel_restr(self, rel):
        nrrel = nodes.Not(_make_relation(self.select, self.filtered_variable,
                                         self.rtype, self.role)[1])
        rel.parent.replace(rel, nodes.Or(nrrel, rel))


class RelationAttributeFacet(RelationFacet):
    """Base facet to filter some entities according to an attribute of other
    entities to which they are related. Most things work similarly as
    :class:`RelationFacet`, except that:

    * `label_vid` doesn't make sense here

    * you should specify the attribute type using `target_attr_type` if it's not a
      String

    * you can specify a comparison operator using `comparator`


    Back to our example... if you want to search office by postal code and that
    you use a :class:`RelationFacet` for that, you won't get the expected
    behaviour: if two offices have the same postal code, they've however two
    different addresses.  So you'll see in the facet the same postal code twice,
    though linked to a different address entity. There is a great chance your
    users won't understand that...

    That's where this class come in! It's used to said that you want to filter
    according to the *attribute value* of a relatied entity, not to the entity
    itself. Now here is the source code for the facet:

    .. sourcecode:: python

      class PostalCodeFacet(RelationAttributeFacet):
          __regid__ = 'postalcode'
          # this facet should only be selected when visualizing offices
          __select__ = RelationAttributeFacet.__select__ & is_instance('Office')
          # this facet is a filter on the PostalAddress entities linked to the
          # office through the 'has_address' relation, where the office is the
          # subject of the relation
          rtype = 'has_address'
          role = 'subject'
          # we want to search according to address 'postal_code' attribute
          target_attr = 'postalcode'
    """
    _select_target_entity = False
    # attribute type
    target_attr_type = 'String'
    # type of comparison: default is an exact match on the attribute value
    comparator = '=' # could be '<', '<=', '>', '>='

    @property
    def restr_attr(self):
        return self.target_attr

    @property
    def restr_attr_type(self):
        return self.target_attr_type

    def rset_vocabulary(self, rset):
        if self.i18nable:
            tr = self._cw._
        else:
            tr = text_type
        if self.rql_sort:
            return [(tr(value), value) for value, in rset]
        values = [(tr(value), value) for value, in rset]
        return sorted(values, reverse=not self.sortasc)


class AttributeFacet(RelationAttributeFacet):
    """Base facet to filter some entities according one of their attribute.
    Configuration is mostly similarly as :class:`RelationAttributeFacet`, except that:

    * `target_attr` doesn't make sense here (you specify the attribute using `rtype`
    * `role` neither, it's systematically 'subject'

    So, suppose that in our office search example you want to refine search according
    to the office's surface. Here is a code snippet achieving this:

    .. sourcecode:: python

      class SurfaceFacet(AttributeFacet):
          __regid__ = 'surface'
          __select__ = AttributeFacet.__select__ & is_instance('Office')
          # this facet is a filter on the office'surface
          rtype = 'surface'
          # override the default value of operator since we want to filter
          # according to a minimal value, not an exact one
          comparator = '>='

          def vocabulary(self):
              '''override the default vocabulary method since we want to
              hard-code our threshold values.

              Not overriding would generate a filter containing all existing
              surfaces defined in the database.
              '''
              return [('> 200', '200'), ('> 250', '250'),
                      ('> 275', '275'), ('> 300', '300')]
    """

    support_and = False
    _select_target_entity = True

    @property
    def i18nable(self):
        """should label be internationalized"""
        for rdef in self._iter_rdefs():
            # no 'internationalizable' property for rdef whose object is not a
            # String
            if not getattr(rdef, 'internationalizable', False):
                return False
        return True

    def vocabulary(self):
        """return vocabulary for this facet, eg a list of 2-uple (label, value)
        """
        select = self.select
        select.save_state()
        try:
            filtered_variable = self.filtered_variable
            cleanup_select(select, filtered_variable)
            newvar = prepare_vocabulary_select(select, filtered_variable, self.rtype, self.role)
            _set_orderby(select, newvar, self.sortasc, self.sortfunc)
            if self.cw_rset:
                args = self.cw_rset.args
            else: # vocabulary used for possible_values
                args = None
            try:
                rset = self.rqlexec(select.as_string(), args)
            except Exception:
                self.exception('error while getting vocabulary for %s, rql: %s',
                               self, select.as_string())
                return ()
        finally:
            select.recover()
        # don't call rset_vocabulary on empty result set, it may be an empty
        # *list* (see rqlexec implementation)
        return rset and self.rset_vocabulary(rset)

    def add_rql_restrictions(self):
        """add restriction for this facet into the rql syntax tree"""
        value = self._cw.form.get(self.__regid__)
        if not value:
            return
        filtered_variable = self.filtered_variable
        self.select.add_constant_restriction(filtered_variable, self.rtype, value,
                                            self.target_attr_type, self.comparator)


class RQLPathFacet(RelationFacet):
    """Base facet to filter some entities according to an arbitrary rql
    path. Path should be specified as a list of 3-uples or triplet string, where
    'X' represent the filtered variable. You should specify using
    `filter_variable` the snippet variable that will be used to filter out
    results. You may also specify a `label_variable`. If you want to filter on
    an attribute value, you usually don't want to specify the later since it's
    the same as the filter variable, though you may have to specify the attribute
    type using `restr_attr_type` if there are some type ambiguity in the schema
    for the attribute.

    Using this facet, we can rewrite facets we defined previously:

    .. sourcecode:: python

      class AgencyFacet(RQLPathFacet):
          __regid__ = 'agency'
          # this facet should only be selected when visualizing offices
          __select__ = is_instance('Office')
          # this facet is a filter on the 'Agency' entities linked to the office
          # through the 'proposed_by' relation, where the office is the subject
          # of the relation
          path = ['X has_address O', 'O name N']
          filter_variable = 'O'
          label_variable = 'N'

      class PostalCodeFacet(RQLPathFacet):
          __regid__ = 'postalcode'
          # this facet should only be selected when visualizing offices
          __select__ = is_instance('Office')
          # this facet is a filter on the PostalAddress entities linked to the
          # office through the 'has_address' relation, where the office is the
          # subject of the relation
          path = ['X has_address O', 'O postal_code PC']
          filter_variable = 'PC'

    Though some features, such as 'no value' or automatic internationalization,
    won't work. This facet class is designed to be used for cases where
    :class:`RelationFacet` or :class:`RelationAttributeFacet` can't do the trick
    (e.g when you want to filter on entities where are not directly linked to
    the filtered entities).
    """
    __select__ = yes() # we don't want RelationFacet's selector
    # must be specified
    path = None
    filter_variable = None
    # may be specified
    label_variable = None
    # usually guessed, but may be explicitly specified
    restr_attr = None
    restr_attr_type = None

    # XXX disabled features
    i18nable = False
    no_relation = False
    support_and = False

    def __init__(self, *args, **kwargs):
        super(RQLPathFacet, self).__init__(*args, **kwargs)
        assert self.filter_variable != self.label_variable, \
            ('filter_variable and label_variable should be different. '
             'You may want to let label_variable undefined (ie None).')
        assert self.path and isinstance(self.path, (list, tuple)), \
            'path should be a list of 3-uples, not %s' % self.path
        for part in self.path:
            if isinstance(part, string_types):
                part = part.split()
            assert len(part) == 3, \
                   'path should be a list of 3-uples, not %s' % part

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__,
                            ','.join(str(p) for p in self.path))

    def vocabulary(self):
        """return vocabulary for this facet, eg a list of (label, value)"""
        select = self.select
        select.save_state()
        if self.rql_sort:
            sort = self.sortasc
        else:
            sort = None # will be sorted on label
        try:
            cleanup_select(select, self.filtered_variable)
            varmap, restrvar = self.add_path_to_select()
            select.append_selected(nodes.VariableRef(restrvar))
            if self.label_variable:
                attrvar = varmap[self.label_variable]
            else:
                attrvar = restrvar
            select.append_selected(nodes.VariableRef(attrvar))
            if sort is not None:
                _set_orderby(select, attrvar, sort, self.sortfunc)
            try:
                rset = self.rqlexec(select.as_string(), self.cw_rset.args)
            except Exception:
                self.exception('error while getting vocabulary for %s, rql: %s',
                               self, select.as_string())
                return ()
        finally:
            select.recover()
        # don't call rset_vocabulary on empty result set, it may be an empty
        # *list* (see rqlexec implementation)
        values = rset and self.rset_vocabulary(rset) or []
        if self._include_no_relation():
            values.insert(0, (self._cw._(self.no_relation_label), ''))
        return values

    def possible_values(self):
        """return a list of possible values (as string since it's used to
        compare to a form value in javascript) for this facet
        """
        select = self.select
        select.save_state()
        try:
            cleanup_select(select, self.filtered_variable)
            varmap, restrvar = self.add_path_to_select(skiplabel=True)
            select.append_selected(nodes.VariableRef(restrvar))
            values = [text_type(x) for x, in self.rqlexec(select.as_string())]
        except Exception:
            self.exception('while computing values for %s', self)
            return []
        finally:
            select.recover()
        if self._include_no_relation():
            values.append('')
        return values

    def add_rql_restrictions(self):
        """add restriction for this facet into the rql syntax tree"""
        value = self._cw.form.get(self.__regid__)
        if value is None:
            return
        varmap, restrvar = self.add_path_to_select(
            skiplabel=True, skipattrfilter=True)
        self.value_restriction(restrvar, None, value)

    def add_path_to_select(self, skiplabel=False, skipattrfilter=False):
        varmap = {'X': self.filtered_variable}
        actual_filter_variable = None
        for part in self.path:
            if isinstance(part, string_types):
                part = part.split()
            subject, rtype, object = part
            if skiplabel and object == self.label_variable:
                continue
            if object == self.filter_variable:
                rschema = self._cw.vreg.schema.rschema(rtype)
                if rschema.final:
                    # filter variable is an attribute variable
                    if self.restr_attr is None:
                        self.restr_attr = rtype
                    if self.restr_attr_type is None:
                        attrtypes = set(obj for subj,obj in rschema.rdefs)
                        if len(attrtypes) > 1:
                            raise Exception('ambigous attribute %s, specify attrtype on %s'
                                            % (rtype, self.__class__))
                        self.restr_attr_type = next(iter(attrtypes))
                    if skipattrfilter:
                        actual_filter_variable = subject
                        continue
            subjvar = _get_var(self.select, subject, varmap)
            objvar = _get_var(self.select, object, varmap)
            rel = nodes.make_relation(subjvar, rtype, (objvar,),
                                      nodes.VariableRef)
            self.select.add_restriction(rel)
        if self.restr_attr is None:
            self.restr_attr = 'eid'
        if self.restr_attr_type is None:
            self.restr_attr_type = 'Int'
        if actual_filter_variable:
            restrvar = varmap[actual_filter_variable]
        else:
            restrvar = varmap[self.filter_variable]
        return varmap, restrvar


class RangeFacet(AttributeFacet):
    """This class allows to filter entities according to an attribute of
    numerical type.

    It displays a slider using `jquery`_ to choose a lower bound and an upper
    bound.

    The example below provides an alternative to the surface facet seen earlier,
    in a more powerful way since

    * lower/upper boundaries are computed according to entities to filter
    * user can specify lower/upper boundaries, not only the lower one

    .. sourcecode:: python

      class SurfaceFacet(RangeFacet):
          __regid__ = 'surface'
          __select__ = RangeFacet.__select__ & is_instance('Office')
          # this facet is a filter on the office'surface
          rtype = 'surface'

    All this with even less code!

    The image below display the rendering of the slider:

    .. image:: ../../images/facet_range.png

    .. _jquery: http://www.jqueryui.com/
    """
    target_attr_type = 'Float' # only numerical types are supported
    needs_update = False # not supported actually

    @property
    def wdgclass(self):
        return FacetRangeWidget

    def _range_rset(self):
        select = self.select
        select.save_state()
        try:
            filtered_variable = self.filtered_variable
            cleanup_select(select, filtered_variable)
            newvar = _add_rtype_relation(select, filtered_variable, self.rtype, self.role)[0]
            minf = nodes.Function('MIN')
            minf.append(nodes.VariableRef(newvar))
            select.add_selected(minf)
            maxf = nodes.Function('MAX')
            maxf.append(nodes.VariableRef(newvar))
            select.add_selected(maxf)
            # add is restriction if necessary
            if filtered_variable.stinfo['typerel'] is None:
                etypes = frozenset(sol[filtered_variable.name] for sol in select.solutions)
                select.add_type_restriction(filtered_variable, etypes)
            try:
                return self.rqlexec(select.as_string(), self.cw_rset.args)
            except Exception:
                self.exception('error while getting vocabulary for %s, rql: %s',
                               self, select.as_string())
                return ()
        finally:
            select.recover()

    def vocabulary(self):
        """return vocabulary for this facet, eg a list of 2-uple (label, value)
        """
        rset = self._range_rset()
        if rset:
            minv, maxv = rset[0]
            return [(text_type(minv), minv), (text_type(maxv), maxv)]
        return []

    def possible_values(self):
        """Return a list of possible values (as string since it's used to
        compare to a form value in javascript) for this facet.
        """
        return [strval for strval, val in self.vocabulary()]

    def get_widget(self):
        """return the widget instance to use to display this facet"""
        values = set(value for _, value in self.vocabulary() if value is not None)
        # Rset with entities (the facet is selected) but without values
        if len(values) < 2:
            return None
        return self.wdgclass(self, min(values), max(values))

    def formatvalue(self, value):
        """format `value` before in order to insert it in the RQL query"""
        return text_type(value)

    def infvalue(self, min=False):
        if min:
            return self._cw.form.get('min_%s_inf' % self.__regid__)
        return self._cw.form.get('%s_inf' % self.__regid__)

    def supvalue(self, max=False):
        if max:
            return self._cw.form.get('max_%s_sup' % self.__regid__)
        return self._cw.form.get('%s_sup' % self.__regid__)

    def add_rql_restrictions(self):
        infvalue = self.infvalue()
        supvalue = self.supvalue()
        if infvalue is None or supvalue is None: # nothing sent
            return
        # when a value is equal to one of the limit, don't add the restriction,
        # else we filter out NULL values implicitly
        if infvalue != self.infvalue(min=True):
            self._add_restriction(infvalue, '>=')
        if supvalue != self.supvalue(max=True):
            self._add_restriction(supvalue, '<=')

    def _add_restriction(self, value, operator):
        self.select.add_constant_restriction(self.filtered_variable,
                                             self.rtype,
                                             self.formatvalue(value),
                                             self.target_attr_type, operator)


class DateRangeFacet(RangeFacet):
    """This class works similarly as the :class:`RangeFacet` but for attribute
    of date type.

    The image below display the rendering of the slider for a date range:

    .. image:: ../../images/facet_date_range.png
    """
    target_attr_type = 'Date' # only date types are supported

    @property
    def wdgclass(self):
        return DateFacetRangeWidget

    def formatvalue(self, value):
        """format `value` before in order to insert it in the RQL query"""
        try:
            date_value = ticks2datetime(float(value))
        except (ValueError, OverflowError):
            return u'"date out-of-range"'
        return '"%s"' % ustrftime(date_value, '%Y/%m/%d')


class AbstractRangeRQLPathFacet(RQLPathFacet):
    """
    The :class:`AbstractRangeRQLPathFacet` is the base class for
    RQLPathFacet-type facets allowing the use of RangeWidgets-like
    widgets (such as (:class:`FacetRangeWidget`,
    class:`DateFacetRangeWidget`) on the parent :class:`RQLPathFacet`
    target attribute.
    """
    __abstract__ = True

    def vocabulary(self):
        """return vocabulary for this facet, eg a list of (label,
        value)"""
        select = self.select
        select.save_state()
        try:
            filtered_variable = self.filtered_variable
            cleanup_select(select, filtered_variable)
            varmap, restrvar = self.add_path_to_select()
            if self.label_variable:
                attrvar = varmap[self.label_variable]
            else:
                attrvar = restrvar
            # start RangeRQLPathFacet
            minf = nodes.Function('MIN')
            minf.append(nodes.VariableRef(restrvar))
            select.add_selected(minf)
            maxf = nodes.Function('MAX')
            maxf.append(nodes.VariableRef(restrvar))
            select.add_selected(maxf)
            # add is restriction if necessary
            if filtered_variable.stinfo['typerel'] is None:
                etypes = frozenset(sol[filtered_variable.name] for sol in select.solutions)
                select.add_type_restriction(filtered_variable, etypes)
            # end RangeRQLPathFacet
            try:
                rset = self.rqlexec(select.as_string(), self.cw_rset.args)
            except Exception:
                self.exception('error while getting vocabulary for %s, rql: %s',
                               self, select.as_string())
                return ()
        finally:
            select.recover()
        # don't call rset_vocabulary on empty result set, it may be an empty
        # *list* (see rqlexec implementation)
        if rset:
            minv, maxv = rset[0]
            return [(text_type(minv), minv), (text_type(maxv), maxv)]
        return []


    def possible_values(self):
        """return a list of possible values (as string since it's used to
        compare to a form value in javascript) for this facet
        """
        return [strval for strval, val in self.vocabulary()]

    def add_rql_restrictions(self):
        infvalue = self.infvalue()
        supvalue = self.supvalue()
        if infvalue is None or supvalue is None: # nothing sent
            return
        varmap, restrvar = self.add_path_to_select(
            skiplabel=True, skipattrfilter=True)
        restrel = None
        for part in self.path:
            if isinstance(part, string_types):
                part = part.split()
            subject, rtype, object = part
            if object == self.filter_variable:
                restrel = rtype
        assert restrel
        # when a value is equal to one of the limit, don't add the restriction,
        # else we filter out NULL values implicitly
        if infvalue != self.infvalue(min=True):

            self._add_restriction(infvalue, '>=', restrvar, restrel)
        if supvalue != self.supvalue(max=True):
            self._add_restriction(supvalue, '<=', restrvar, restrel)

    def _add_restriction(self, value, operator, restrvar, restrel):
        self.select.add_constant_restriction(restrvar,
                                             restrel,
                                             self.formatvalue(value),
                                             self.target_attr_type, operator)


class RangeRQLPathFacet(AbstractRangeRQLPathFacet, RQLPathFacet):
    """
    The :class:`RangeRQLPathFacet` uses the :class:`FacetRangeWidget`
    on the :class:`AbstractRangeRQLPathFacet` target attribute
    """
    pass


class DateRangeRQLPathFacet(AbstractRangeRQLPathFacet, DateRangeFacet):
    """
    The :class:`DateRangeRQLPathFacet` uses the
    :class:`DateFacetRangeWidget` on the
    :class:`AbstractRangeRQLPathFacet` target attribute
    """
    pass


class HasRelationFacet(AbstractFacet):
    """This class simply filter according to the presence of a relation
    (whatever the entity at the other end). It display a simple checkbox that
    lets you refine your selection in order to get only entities that actually
    have this relation. You simply have to define which relation using the
    `rtype` and `role` attributes.

    Here is an example of the rendering of thos facet to filter book with image
    and the corresponding code:

    .. image:: ../../images/facet_has_image.png

    .. sourcecode:: python

      class HasImageFacet(HasRelationFacet):
          __regid__ = 'hasimage'
          __select__ = HasRelationFacet.__select__ & is_instance('Book')
          rtype = 'has_image'
          role = 'subject'
    """
    __select__ = partial_relation_possible() & match_context_prop()
    rtype = None # override me in subclass
    role = 'subject' # role of filtered entity in the relation

    title = property(rtype_facet_title)
    needs_update = False # not supported actually
    support_and = False

    def get_widget(self):
        return CheckBoxFacetWidget(self._cw, self,
                                   '%s:%s' % (self.rtype, self),
                                   self._cw.form.get(self.__regid__))

    def add_rql_restrictions(self):
        """add restriction for this facet into the rql syntax tree"""
        value = self._cw.form.get(self.__regid__)
        if not value: # no value sent for this facet
            return
        exists = nodes.Exists()
        self.select.add_restriction(exists)
        var = self.select.make_variable()
        if self.role == 'subject':
            subj, obj = self.filtered_variable, var
        else:
            subj, obj = var, self.filtered_variable
        exists.add_relation(subj, self.rtype, obj)


class BitFieldFacet(AttributeFacet):
    """Base facet class for Int field holding some bit values using binary
    masks.

    label / value for each bit should be given using the :attr:`choices`
    attribute.

    See also :class:`~cubicweb.web.formwidgets.BitSelect`.
    """
    choices = None # to be set on concret class
    def add_rql_restrictions(self):
        value = self._cw.form.get(self.__regid__)
        if not value:
            return
        if isinstance(value, list):
            value = reduce(lambda x, y: int(x) | int(y), value)
        else:
            value = int(value)
        attr_var = self.select.make_variable()
        self.select.add_relation(self.filtered_variable, self.rtype, attr_var)
        comp = nodes.Comparison('=', nodes.Constant(value, 'Int'))
        if value == 0:
            comp.append(nodes.variable_ref(attr_var))
        else:
            comp.append(nodes.MathExpression('&', nodes.variable_ref(attr_var),
                                             nodes.Constant(value, 'Int')))
        having = self.select.having
        if having:
            self.select.replace(having[0], nodes.And(having[0], comp))
        else:
            self.select.set_having([comp])

    def rset_vocabulary(self, rset):
        mask = reduce(lambda x, y: x | (y[0] or 0), rset, 0)
        return sorted([(self._cw._(label), val) for label, val in self.choices
                       if not val or val & mask])

    def possible_values(self):
        return [text_type(val) for label, val in self.vocabulary()]


## html widets ################################################################
_DEFAULT_VOCAB_WIDGET_HEIGHT = 12
_DEFAULT_FACET_GROUP_HEIGHT = 15

class FacetVocabularyWidget(htmlwidgets.HTMLWidget):

    def __init__(self, facet):
        self.facet = facet
        self.items = []

    @cachedproperty
    def css_overflow_limit(self):
        """ we try to deduce a number of displayed lines from a css property
        if we get another unit we're out of luck and resort to one constant
        hence, it is strongly advised not to specify but ems for this css prop
        """
        return css_em_num_value(self.facet._cw.vreg, 'facet_vocabMaxHeight',
                                _DEFAULT_VOCAB_WIDGET_HEIGHT)

    @cachedproperty
    def height(self):
        """ title, optional and/or dropdown, len(items) or upper limit """
        return (1.5 + # title + small magic constant
                int(self.facet._support_and_compat() +
                    min(len(self.items), self.css_overflow_limit)))

    @property
    @cached
    def overflows(self):
        return len(self.items) >= self.css_overflow_limit

    scrollbar_padding_factor = 4

    def _render(self):
        w = self.w
        title = xml_escape(self.facet.title)
        facetid = domid(make_uid(self.facet.__regid__))
        w(u'<div id="%s" class="facet">\n' % facetid)
        cssclass = 'facetTitle'
        if self.facet.allow_hide:
            cssclass += ' hideFacetBody'
        w(u'<div class="%s" cubicweb:facetName="%s">%s</div>\n' %
          (cssclass, xml_escape(self.facet.__regid__), title))
        if self.facet._support_and_compat():
            self._render_and_or(w)
        cssclass = 'facetBody vocabularyFacet'
        if not self.facet.start_unfolded:
            cssclass += ' hidden'
        overflow = self.overflows
        if overflow:
            if self.facet._support_and_compat():
                cssclass += ' vocabularyFacetBodyWithLogicalSelector'
            else:
                cssclass += ' vocabularyFacetBody'
        w(u'<div class="%s">\n' % cssclass)
        for value, label, selected in self.items:
            if value is None:
                continue
            self._render_value(w, value, label, selected, overflow)
        w(u'</div>\n')
        w(u'</div>\n')

    def _render_and_or(self, w):
        _ = self.facet._cw._
        w(u"""<select name='%s' class='radio facetOperator' title='%s'>
  <option value='OR'>%s</option>
  <option value='AND'>%s</option>
</select>""" % (xml_escape(self.facet.__regid__) + '_andor',
                _('and/or between different values'),
                _('OR'), _('AND')))

    def _render_value(self, w, value, label, selected, overflow):
        cssclass = 'facetValue facetCheckBox'
        if selected:
            cssclass += ' facetValueSelected'
        w(u'<div class="%s" cubicweb:value="%s">\n'
          % (cssclass, xml_escape(text_type(value))))
        # If it is overflowed one must add padding to compensate for the vertical
        # scrollbar; given current css values, 4 blanks work perfectly ...
        padding = u'&#160;' * self.scrollbar_padding_factor if overflow else u''
        w('<span>%s</span>' % xml_escape(label))
        w(padding)
        w(u'</div>')

class FacetStringWidget(htmlwidgets.HTMLWidget):
    def __init__(self, facet):
        self.facet = facet
        self.value = None

    @property
    def height(self):
        return 2.5

    def _render(self):
        w = self.w
        title = xml_escape(self.facet.title)
        facetid = make_uid(self.facet.__regid__)
        w(u'<div id="%s" class="facet">\n' % facetid)
        cssclass = 'facetTitle'
        if self.facet.allow_hide:
            cssclass += ' hideFacetBody'
        w(u'<div class="%s" cubicweb:facetName="%s">%s</div>\n' %
               (cssclass, xml_escape(self.facet.__regid__), title))
        cssclass = 'facetBody'
        if not self.facet.start_unfolded:
            cssclass += ' hidden'
        w(u'<div class="%s">\n' % cssclass)
        w(u'<input name="%s" type="text" value="%s" />\n' % (
                xml_escape(self.facet.__regid__), self.value or u''))
        w(u'</div>\n')
        w(u'</div>\n')


class FacetRangeWidget(htmlwidgets.HTMLWidget):
    formatter = 'function (value) {return value;}'
    onload = u'''
    var _formatter = %(formatter)s;
    jQuery("#%(sliderid)s").slider({
        range: true,
        min: %(minvalue)s,
        max: %(maxvalue)s,
        values: [%(minvalue)s, %(maxvalue)s],
        stop: function(event, ui) { // submit when the user stops sliding
           var form = $('#%(sliderid)s').closest('form');
           buildRQL.apply(null, cw.evalJSON(form.attr('cubicweb:facetargs')));
        },
        slide: function(event, ui) {
            jQuery('#%(sliderid)s_inf').html(_formatter(ui.values[0]));
            jQuery('#%(sliderid)s_sup').html(_formatter(ui.values[1]));
            jQuery('input[name="%(facetname)s_inf"]').val(ui.values[0]);
            jQuery('input[name="%(facetname)s_sup"]').val(ui.values[1]);
        }
   });
   // use JS formatter to format value on page load
   jQuery('#%(sliderid)s_inf').html(_formatter(jQuery('input[name="%(facetname)s_inf"]').val()));
   jQuery('#%(sliderid)s_sup').html(_formatter(jQuery('input[name="%(facetname)s_sup"]').val()));
'''
    #'# make emacs happier
    def __init__(self, facet, minvalue, maxvalue):
        self.facet = facet
        self.minvalue = minvalue
        self.maxvalue = maxvalue

    @property
    def height(self):
        return 2.5

    def _render(self):
        w = self.w
        facet = self.facet
        facet._cw.add_js('jquery.ui.js')
        facet._cw.add_css('jquery.ui.css')
        sliderid = make_uid('theslider')
        facetname = self.facet.__regid__
        facetid = make_uid(facetname)
        facet._cw.html_headers.add_onload(self.onload % {
            'sliderid': sliderid,
            'facetid': facetid,
            'facetname': facetname,
            'minvalue': self.minvalue,
            'maxvalue': self.maxvalue,
            'formatter': self.formatter,
            })
        title = xml_escape(self.facet.title)
        facetname = xml_escape(facetname)
        w(u'<div id="%s" class="facet rangeFacet">\n' % facetid)
        cssclass = 'facetTitle'
        if facet.allow_hide:
            cssclass += ' hideFacetBody'
        w(u'<div class="%s" cubicweb:facetName="%s">%s</div>\n' %
          (cssclass, facetname, title))
        cssclass = 'facetBody'
        if not self.facet.start_unfolded:
            cssclass += ' hidden'
        w(u'<div class="%s">\n' % cssclass)
        w(u'<span id="%s_inf"></span> - <span id="%s_sup"></span>'
          % (sliderid, sliderid))
        w(u'<input type="hidden" name="%s_inf" value="%s" />'
          % (facetname, self.minvalue))
        w(u'<input type="hidden" name="%s_sup" value="%s" />'
          % (facetname, self.maxvalue))
        w(u'<input type="hidden" name="min_%s_inf" value="%s" />'
          % (facetname, self.minvalue))
        w(u'<input type="hidden" name="max_%s_sup" value="%s" />'
          % (facetname, self.maxvalue))
        w(u'<div id="%s"></div>' % sliderid)
        w(u'</div>\n')
        w(u'</div>\n')


class DateFacetRangeWidget(FacetRangeWidget):

    formatter = 'function (value) {return (new Date(parseFloat(value))).strftime(DATE_FMT);}'

    def round_max_value(self, d):
        'round to upper value to avoid filtering out the max value'
        return datetime(d.year, d.month, d.day) + timedelta(days=1)

    def __init__(self, facet, minvalue, maxvalue):
        maxvalue = self.round_max_value(maxvalue)
        super(DateFacetRangeWidget, self).__init__(facet,
                                                   datetime2ticks(minvalue),
                                                   datetime2ticks(maxvalue))
        fmt = facet._cw.property_value('ui.date-format')
        facet._cw.html_headers.define_var('DATE_FMT', fmt)


class CheckBoxFacetWidget(htmlwidgets.HTMLWidget):
    selected_img = "black-check.png"
    unselected_img = "black-uncheck.png"

    def __init__(self, req, facet, value, selected):
        self._cw = req
        self.facet = facet
        self.value = value
        self.selected = selected

    @property
    def height(self):
        return 1.5

    def _render(self):
        w = self.w
        title = xml_escape(self.facet.title)
        facetid = make_uid(self.facet.__regid__)
        w(u'<div id="%s" class="facet">\n' % facetid)
        cssclass = 'facetValue facetCheckBox'
        if self.selected:
            cssclass += ' facetValueSelected'
            imgsrc = self._cw.data_url(self.selected_img)
            imgalt = self._cw._('selected')
        else:
            imgsrc = self._cw.data_url(self.unselected_img)
            imgalt = self._cw._('not selected')
        w(u'<div class="%s" cubicweb:value="%s">\n'
          % (cssclass, xml_escape(text_type(self.value))))
        w(u'<div>')
        w(u'<img src="%s" alt="%s" cubicweb:unselimg="true" />&#160;' % (imgsrc, imgalt))
        w(u'<label class="facetTitle" cubicweb:facetName="%s">%s</label>'
          % (xml_escape(self.facet.__regid__), title))
        w(u'</div>\n')
        w(u'</div>\n')
        w(u'</div>\n')


# other classes ################################################################

class FilterRQLBuilder(object):
    """called by javascript to get a rql string from filter form"""

    def __init__(self, req):
        self._cw = req

    def build_rql(self):
        form = self._cw.form
        facetids = form['facets'].split(',')
        # XXX Union unsupported yet
        select = self._cw.vreg.parse(self._cw, form['baserql']).children[0]
        filtered_variable = get_filtered_variable(select, form.get('mainvar'))
        toupdate = []
        for facetid in facetids:
            facet = get_facet(self._cw, facetid, select, filtered_variable)
            facet.add_rql_restrictions()
            if facet.needs_update:
                toupdate.append(facetid)
        return select.as_string(), toupdate



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
import copy
import warnings

from logilab.common.testlib import tag
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.web import uihelper, formwidgets as fwdgs
from cubicweb.web.views import uicfg

abaa = uicfg.actionbox_appearsin_addmenu

class UICFGTC(CubicWebTC):

    def test_default_actionbox_appearsin_addmenu_config(self):
        self.assertFalse(abaa.etype_get('TrInfo', 'wf_info_for', 'object', 'CWUser'))



class DefinitionOrderTC(CubicWebTC):
    """This test check that when multiple definition could match a key, only
    the more accurate apply"""

    def setUp(self):
        super(DefinitionOrderTC, self).setUp()
        for rtag in (uicfg.autoform_section, uicfg.autoform_field_kwargs):
            rtag._old_tagdefs = copy.deepcopy(rtag._tagdefs)
        new_def = (
                    (('*', 'login', '*'),
                         {'formtype':'main', 'section':'hidden'}),
                    (('*', 'login', '*'),
                         {'formtype':'muledit', 'section':'hidden'}),
                    (('CWUser', 'login', '*'),
                         {'formtype':'main', 'section':'attributes'}),
                    (('CWUser', 'login', '*'),
                         {'formtype':'muledit', 'section':'attributes'}),
                    (('CWUser', 'login', 'String'),
                         {'formtype':'main', 'section':'inlined'}),
                    (('CWUser', 'login', 'String'),
                         {'formtype':'inlined', 'section':'attributes'}),
                    )
        for key, kwargs in new_def:
            uicfg.autoform_section.tag_subject_of(key, **kwargs)

    def tearDown(self):
        super(DefinitionOrderTC, self).tearDown()
        for rtag in (uicfg.autoform_section, uicfg.autoform_field_kwargs):
            rtag._tagdefs = rtag._old_tagdefs

    @tag('uicfg')
    def test_definition_order_hidden(self):
        result = uicfg.autoform_section.get('CWUser', 'login', 'String', 'subject')
        expected = set(['main_inlined', 'muledit_attributes', 'inlined_attributes'])
        self.assertSetEqual(result, expected)

    @tag('uihelper', 'order', 'func')
    def test_uihelper_set_fields_order(self):
        afk_get = uicfg.autoform_field_kwargs.get
        self.assertEqual(afk_get('CWUser', 'firstname', 'String', 'subject'), {})
        with warnings.catch_warnings(record=True) as w:
            uihelper.set_fields_order('CWUser', ('login', 'firstname', 'surname'))
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
        self.assertEqual(afk_get('CWUser', 'firstname', 'String', 'subject'), {'order': 1})

    @tag('uicfg', 'order', 'func')
    def test_uicfg_primaryview_set_fields_order(self):
        pvdc = uicfg.primaryview_display_ctrl
        pvdc.set_fields_order('CWUser', ('login', 'firstname', 'surname'))
        self.assertEqual(pvdc.get('CWUser', 'login', 'String', 'subject'), {'order': 0})
        self.assertEqual(pvdc.get('CWUser', 'firstname', 'String', 'subject'), {'order': 1})
        self.assertEqual(pvdc.get('CWUser', 'surname', 'String', 'subject'), {'order': 2})

    @tag('uihelper', 'kwargs', 'func')
    def test_uihelper_set_field_kwargs(self):
        afk_get = uicfg.autoform_field_kwargs.get
        self.assertEqual(afk_get('CWUser', 'firstname', 'String', 'subject'), {})
        wdg = fwdgs.TextInput({'size': 30})
        with warnings.catch_warnings(record=True) as w:
            uihelper.set_field_kwargs('CWUser', 'firstname', widget=wdg)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
        self.assertEqual(afk_get('CWUser', 'firstname', 'String', 'subject'), {'widget': wdg})

    @tag('uihelper', 'hidden', 'func')
    def test_uihelper_hide_fields(self):
        # original conf : in_group is edited in 'attributes' section everywhere
        section_conf = uicfg.autoform_section.get('CWUser', 'in_group', '*', 'subject')
        self.assertCountEqual(section_conf, ['main_attributes', 'muledit_attributes'])
        # hide field in main form
        with warnings.catch_warnings(record=True) as w:
            uihelper.hide_fields('CWUser', ('login', 'in_group'))
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
        section_conf = uicfg.autoform_section.get('CWUser', 'in_group', '*', 'subject')
        self.assertCountEqual(section_conf, ['main_hidden', 'muledit_attributes'])
        # hide field in muledit form
        with warnings.catch_warnings(record=True) as w:
            uihelper.hide_fields('CWUser', ('login', 'in_group'), formtype='muledit')
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
        section_conf = uicfg.autoform_section.get('CWUser', 'in_group', '*', 'subject')
        self.assertCountEqual(section_conf, ['main_hidden', 'muledit_hidden'])

    @tag('uihelper', 'hidden', 'formconfig')
    def test_uihelper_formconfig(self):
        afk_get = uicfg.autoform_field_kwargs.get
        class CWUserFormConfig(uihelper.FormConfig):
            etype = 'CWUser'
            hidden = ('in_group',)
            fields_order = ('login', 'firstname')
        section_conf = uicfg.autoform_section.get('CWUser', 'in_group', '*', 'subject')
        self.assertCountEqual(section_conf, ['main_hidden', 'muledit_attributes'])
        self.assertEqual(afk_get('CWUser', 'firstname', 'String', 'subject'), {'order': 1})


class UicfgRegistryTC(CubicWebTC):

    def test_default_uicfg_object(self):
        'CW default ui config objects must be registered in uicfg registry'
        onames = ('autoform_field', 'autoform_section', 'autoform_field_kwargs')
        for oname in onames:
            obj = self.vreg['uicfg'].select_or_none(oname)
            self.assertTrue(obj is not None, '%s not found in uicfg registry'
                            % oname)

    def test_custom_uicfg(self):
        ASRT = uicfg.AutoformSectionRelationTags
        custom_afs = ASRT()
        custom_afs.__select__ = ASRT.__select__ & ASRT.__select__
        self.vreg['uicfg'].register(custom_afs)
        obj = self.vreg['uicfg'].select_or_none('autoform_section')
        self.assertTrue(obj is custom_afs)


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# -*- coding: iso-8859-1 -*-
# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unit tests for modules cubicweb.server.rqlannotation"""

from cubicweb import devtools
from cubicweb.devtools.repotest import BaseQuerierTC


class SQLGenAnnotatorTC(BaseQuerierTC):

    def setUp(self):
        handler = devtools.get_test_db_handler(devtools.TestServerConfiguration('data', __file__))
        handler.build_db_cache()
        repo, _cnx = handler.get_repo_and_cnx()
        self.__class__.repo = repo
        super(SQLGenAnnotatorTC, self).setUp()

    def get_max_eid(self):
        # no need for cleanup here
        return None

    def cleanup(self):
        # no need for cleanup here
        pass

    def test_0_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any SEN,RN,OEN WHERE X from_entity SE, '
                                  'SE eid 44, X relation_type R, R eid 139, '
                                  'X to_entity OE, OE eid 42, R name RN, SE name SEN, '
                                  'OE name OEN')
            self.assertEqual(rqlst.defined_vars['SE']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['OE']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['R']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['SE'].stinfo['attrvar'], None)
            self.assertEqual(rqlst.defined_vars['OE'].stinfo['attrvar'], None)
            self.assertEqual(rqlst.defined_vars['R'].stinfo['attrvar'], None)

    def test_0_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any O WHERE NOT S ecrit_par O, S eid 1, '
                                  'S inline1 P, O inline2 P')
            self.assertEqual(rqlst.defined_vars['P']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['O'].stinfo['attrvar'], None)

    def test_0_4(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any A,B,C WHERE A eid 12,A comment B, '
                                  'A ?wf_info_for C')
            self.assertEqual(rqlst.defined_vars['A']._q_invariant, False)
            self.assertTrue(rqlst.defined_vars['B'].stinfo['attrvar'])
            self.assertEqual(rqlst.defined_vars['C']._q_invariant, False)
            self.assertEqual(rqlst.solutions, [{'A': 'TrInfo', 'B': 'String', 'C': 'Affaire'},
                                               {'A': 'TrInfo', 'B': 'String', 'C': 'CWUser'},
                                               {'A': 'TrInfo', 'B': 'String', 'C': 'Note'}])

    def test_0_5(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any P WHERE N ecrit_par P, N eid 0')
            self.assertEqual(rqlst.defined_vars['N']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['P']._q_invariant, True)

    def test_0_6(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any P WHERE NOT N ecrit_par P, N eid 512')
            self.assertEqual(rqlst.defined_vars['P']._q_invariant, False)

    def test_0_7(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Personne X,Y where X nom NX, '
                                  'Y nom NX, X eid XE, not Y eid XE')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)
            self.assertTrue(rqlst.defined_vars['XE'].stinfo['attrvar'])

    def test_0_8(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any P WHERE X eid 0, NOT X connait P')
            self.assertEqual(rqlst.defined_vars['P']._q_invariant, False)
            self.assertEqual(len(rqlst.solutions), 1, rqlst.solutions)

    def test_0_10(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X concerne Y, Y is Note')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)

    def test_0_11(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X todo_by Y, X is Affaire')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, True)

    def test_0_12(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Personne P WHERE P concerne A, '
                                  'A concerne S, S nom "Logilab"')
            self.assertEqual(rqlst.defined_vars['P']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['A']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['S']._q_invariant, False)

    def test_1_0(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X,Y WHERE X created_by Y, '
                                  'X eid 5, NOT Y eid 6')
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, True)

    def test_1_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X,Y WHERE X created_by Y, X eid 5, '
                                  'NOT Y eid IN (6,7)')
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, True)

    def test_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X identity Y, Y eid 1')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)

    def test_7(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Personne X,Y where X nom NX, Y nom NX, '
                                  'X eid XE, not Y eid XE')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)

    def test_8(self):
        with self.admin_access.cnx() as cnx:
            # DISTINCT Any P WHERE P require_group %(g)s,
            # NOT %(u)s has_group_permission P, P is CWPermission
            rqlst = self._prepare(cnx, 'DISTINCT Any X WHERE A concerne X, '
                                  'NOT N migrated_from X, '
                                  'X is Note, N eid 1')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)

    def test_diff_scope_identity_deamb(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X concerne Y, Y is Note, '
                                  'EXISTS(Y identity Z, Z migrated_from N)')
            self.assertEqual(rqlst.defined_vars['Z']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, True)

    def test_optional_inlined(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X,S where X from_state S?')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['S']._q_invariant, True)

    def test_optional_inlined_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any N,A WHERE N? inline1 A')
            self.assertEqual(rqlst.defined_vars['N']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['A']._q_invariant, False)

    def test_optional_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X,S WHERE X travaille S?')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['S']._q_invariant, True)

    def test_greater_eid(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X eid > 5')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)

    def test_greater_eid_typed(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X eid > 5, X is Note')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)

    def test_max_eid(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any MAX(X)')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)

    def test_max_eid_typed(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any MAX(X) WHERE X is Note')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)

    def test_all_entities(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)

    def test_all_typed_entity(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X is Note')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)

    def test_has_text_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X has_text "toto tata"')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['X'].stinfo['principal'].r_type,
                             'has_text')

    def test_has_text_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X is Personne, '
                                  'X has_text "coucou"')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['X'].stinfo['principal'].r_type,
                             'has_text')

    def test_not_relation_1(self):
        with self.admin_access.cnx() as cnx:
            # P can't be invariant since deambiguification caused by "NOT X require_permission P"
            # is not considered by generated sql (NOT EXISTS(...))
            rqlst = self._prepare(cnx, 'Any P,G WHERE P require_group G, '
                                  'NOT X require_permission P')
            self.assertEqual(rqlst.defined_vars['P']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['G']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)

    def test_not_relation_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'TrInfo X WHERE X eid 2, '
                                  'NOT X from_state Y, Y is State')
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)

    def test_not_relation_3(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X, Y WHERE X eid 1, Y eid in (2, 3)')
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)

    def test_not_relation_4_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Note X WHERE NOT Y evaluee X')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, True)

    def test_not_relation_4_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE NOT Y evaluee X')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, True)

    def test_not_relation_4_3(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any Y WHERE NOT Y evaluee X')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)

    def test_not_relation_4_4(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE NOT Y evaluee X, Y is CWUser')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)

    def test_not_relation_4_5(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE NOT Y evaluee X, '
                                  'Y eid %s, X is Note' % self.ueid)
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.solutions, [{'X': 'Note'}])

    def test_not_relation_5_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X,Y WHERE X name "CWGroup", '
                                  'Y eid IN(1, 2, 3), NOT X read_permission Y')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)

    def test_not_relation_5_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'DISTINCT Any X,Y WHERE X name "CWGroup", '
                                  'Y eid IN(1, 2, 3), NOT X read_permission Y')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)

    def test_not_relation_6(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Personne P where NOT P concerne A')
            self.assertEqual(rqlst.defined_vars['P']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['A']._q_invariant, True)

    def test_not_relation_7(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any K,V WHERE P is CWProperty, '
                                  'P pkey K, P value V, NOT P for_user U')
            self.assertEqual(rqlst.defined_vars['P']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['U']._q_invariant, True)

    def test_exists_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any U WHERE U eid IN (1,2), EXISTS(X owned_by U)')
            self.assertEqual(rqlst.defined_vars['U']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)

    def test_exists_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any U WHERE EXISTS(U eid IN (1,2), X owned_by U)')
            self.assertEqual(rqlst.defined_vars['U']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)

    def test_exists_3(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any U WHERE EXISTS(X owned_by U, X bookmarked_by U)')
            self.assertEqual(rqlst.defined_vars['U']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)

    def test_exists_4(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X,Y WHERE X name "CWGroup", '
                                  'Y eid IN(1, 2, 3), EXISTS(X read_permission Y)')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)

    def test_exists_5(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'DISTINCT Any X,Y WHERE X name "CWGroup", '
                                  'Y eid IN(1, 2, 3), EXISTS(X read_permission Y)')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, True)

    def test_not_exists_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any U WHERE NOT EXISTS(X owned_by U, '
                                  'X bookmarked_by U)')
            self.assertEqual(rqlst.defined_vars['U']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)

    def test_not_exists_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X,Y WHERE X name "CWGroup", '
                                  'Y eid IN(1, 2, 3), NOT EXISTS(X read_permission Y)')
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)

    def test_not_exists_distinct_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'DISTINCT Any X,Y WHERE X name "CWGroup", '
                                  'Y eid IN(1, 2, 3), NOT EXISTS(X read_permission Y)')
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, False)

    def test_or_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X concerne B OR '
                                  'C concerne X, B eid 12, C eid 13')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)

    def test_or_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X created_by U, X concerne B OR '
                                  'C concerne X, B eid 12, C eid 13')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['U']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['X'].stinfo['principal'].r_type, 'created_by')

    def test_or_3(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any N WHERE A evaluee N or EXISTS(N todo_by U)')
            self.assertEqual(rqlst.defined_vars['N']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['A']._q_invariant, True)
            self.assertEqual(rqlst.defined_vars['U']._q_invariant, True)

    def test_or_exists_1(self):
        with self.admin_access.cnx() as cnx:
            # query generated by security rewriting
            rqlst = self._prepare(cnx, 'DISTINCT Any A,S WHERE A is Affaire, S nom "chouette", '
                                  'S is IN(Division, Societe, SubDivision),'
                                  '(EXISTS(A owned_by D)) '
                                  'OR ((((EXISTS(E concerne C?, C owned_by D, A identity E, '
                                  '              C is Note, E is Affaire)) '
                                  'OR (EXISTS(I concerne H?, H owned_by D, H is Societe, '
                                  '           A identity I, I is Affaire))) '
                                  'OR (EXISTS(J concerne G?, G owned_by D, G is SubDivision, '
                                  '           A identity J, J is Affaire))) '
                                  'OR (EXISTS(K concerne F?, F owned_by D, F is Division, '
                                  '           A identity K, K is Affaire)))')
            self.assertEqual(rqlst.defined_vars['A']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['S']._q_invariant, False)

    def test_or_exists_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any U WHERE EXISTS(U in_group G, G name "managers") OR '
                                  'EXISTS(X owned_by U, X bookmarked_by U)')
            self.assertEqual(rqlst.defined_vars['U']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['G']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, True)

    def test_or_exists_3(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any COUNT(S),CS GROUPBY CS ORDERBY 1 DESC LIMIT 10 '
                                  'WHERE C is Societe, S concerne C, C nom CS, '
                                  '(EXISTS(S owned_by D)) '
                                  'OR (EXISTS(S documented_by N, N title "published"))')
            self.assertEqual(rqlst.defined_vars['S']._q_invariant, True)
            rqlst = self._prepare(cnx, 'Any COUNT(S),CS GROUPBY CS ORDERBY 1 DESC LIMIT 10 '
                                  'WHERE S is Affaire, C is Societe, S concerne C, C nom CS, '
                                  '(EXISTS(S owned_by D)) '
                                  'OR (EXISTS(S documented_by N, N title "published"))')
            self.assertEqual(rqlst.defined_vars['S']._q_invariant, True)

    def test_nonregr_ambiguity(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Note N WHERE N attachment F')
            # N may be an image as well, not invariant
            self.assertEqual(rqlst.defined_vars['N']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['F']._q_invariant, True)

    def test_nonregr_ambiguity_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any S,SN WHERE X has_text "tot", '
                                  'X in_state S, S name SN, X is CWUser')
            # X use has_text but should not be invariant as ambiguous, and has_text
            # may not be its principal
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['S']._q_invariant, False)

    def test_remove_from_deleted_source_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Note X WHERE X eid 999998, NOT X cw_source Y')
            self.assertNotIn('X', rqlst.defined_vars)  # simplified
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, True)

    def test_remove_from_deleted_source_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Note X WHERE X eid IN (999998, 999999), NOT X cw_source Y')
            self.assertEqual(rqlst.defined_vars['X']._q_invariant, False)
            self.assertEqual(rqlst.defined_vars['Y']._q_invariant, True)

    def test_has_text_security_cache_bug(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X has_text "toto" WITH X BEING '
                                  '(Any C WHERE C is Societe, C nom CS)')
            self.assertTrue(rqlst.parent.has_text_query)


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Basic views for python values (eg without any result set)
"""


from six import text_type
from six.moves import range

from cubicweb.view import View
from cubicweb.predicates import match_kwargs
from cubicweb.web.views import tableview


class PyValTableColRenderer(tableview.AbstractColumnRenderer):
    """Default column renderer for :class:`PyValTableView`."""
    def bind(self, view, colid):
        super(PyValTableColRenderer, self).bind(view, colid)
        self.header = view.headers[colid] if view.headers else None
        self.data = view.pyvalue

    def render_header(self, w):
        if self.header:
            w(self._cw._(self.header))
        else:
            w(self.empty_cell_content)

    def render_cell(self, w, rownum):
        w(text_type(self.data[rownum][self.colid]))


class PyValTableView(tableview.TableMixIn, View):
    """This table view is designed to be used a list of list of unicode values
    given as a mandatory `pyvalue` argument. Take care, content is NOT
    xml-escaped.

    It's configured through the following selection arguments.

    If `headers` is specified, it is expected to be a list of headers to be
    inserted as first row (in <thead>).

    `header_column_idx` may be used to specify a column index or a set of column
    indiced where values should be inserted inside <th> tag instead of <td>.

    `cssclass` is the CSS class used on the <table> tag, and default to
    'listing' (so that the table will look similar to those generated by the
    table view).
    """
    __regid__ = 'pyvaltable'
    __select__ = match_kwargs('pyvalue')
    default_column_renderer_class = PyValTableColRenderer
    paginable = False # not supported
    headers = None
    cssclass = None
    domid = None

    def __init__(self, req, pyvalue, headers=None, cssclass=None,
                 header_column_idx=None, **kwargs):
        super(PyValTableView, self).__init__(req, **kwargs)
        self.pyvalue = pyvalue
        if headers is not None:
            self.headers = headers
        elif self.headers: # headers set on a class attribute, translate
            self.headers = [self._cw._(header) for header in self.headers]
        if cssclass is not None:
            self.cssclass = cssclass
        self.header_column_idx = header_column_idx

    @property
    def layout_args(self):
        args = {}
        if self.cssclass:
            args['cssclass'] = self.cssclass
        if self.header_column_idx is not None:
            args['header_column_idx'] = self.header_column_idx
        return args

    # layout callbacks #########################################################

    @property
    def table_size(self):
        """return the number of rows (header excluded) to be displayed"""
        return len(self.pyvalue)

    @property
    def has_headers(self):
        return self.headers

    def build_column_renderers(self):
        return [self.column_renderer(colid)
                for colid in range(len(self.pyvalue[0]))]

    def facets_form(self, mainvar=None):
        return None # not supported

    def table_actions(self):
        return [] # not supported


class PyValListView(View):
    """display a list of values into an html list.

    Take care, content is NOT xml-escaped.
    """
    __regid__ = 'pyvallist'
    __select__ = match_kwargs('pyvalue')

    def call(self, pyvalue):
        self.w(u'<ul>\n')
        for line in pyvalue:
            self.w(u'<li>%s</li>\n' % line)
        self.w(u'</ul>\n')



# copyright 2014-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

from logilab.common import tempattr, attrdict

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.web.views.autoform import InlinedFormField
from cubicweb.web.views.forms import EntityFieldsForm


class InlinedFormTC(CubicWebTC):

    def test_linked_to(self):
        with self.admin_access.web_request() as req:
            formview = req.vreg['views'].select(
                'inline-creation', req,
                etype='File', rtype='described_by_test', role='subject',
                peid=123,
                petype='Salesterm')
            self.assertEqual({('described_by_test', 'object'): [123]},
                             formview.form.linked_to)

    def test_linked_to_parent_being_created(self):
        with self.admin_access.web_request() as req:
            formview = req.vreg['views'].select(
                'inline-creation', req,
                etype='File', rtype='described_by_test', role='subject',
                peid='A',
                petype='Salesterm')
            self.assertEqual(formview.form.linked_to, {})

    def test_remove_js_depending_on_cardinality(self):
        with self.admin_access.web_request() as req:
            formview = req.vreg['views'].select(
                'inline-creation', req,
                etype='File', rtype='described_by_test', role='subject',
                peid='A',
                petype='Salesterm')
            # cardinality is 1, can't remove
            self.assertIsNone(formview._get_removejs())
            rdef = self.schema['Salesterm'].rdef('described_by_test')
            with tempattr(rdef, 'cardinality', '?*'):
                self.assertTrue(formview._get_removejs())
            with tempattr(rdef, 'cardinality', '+*'):
                # formview has no parent info (pform). This is what happens
                # when an inline form is requested through AJAX.
                self.assertTrue(formview._get_removejs())
                fakeview = attrdict(dict(rtype='described_by_test', role='subject'))
                # formview is first, can't be removed
                formview.pform = attrdict(fields=[InlinedFormField(view=formview),
                                                  InlinedFormField(view=fakeview)])
                self.assertIsNone(formview._get_removejs())
                # formview isn't first, can be removed
                formview.pform = attrdict(fields=[InlinedFormField(view=fakeview),
                                                  InlinedFormField(view=formview)])
                self.assertTrue(formview._get_removejs())

    def test_field_by_name_consider_aff(self):
        class MyField(object):
            def __init__(self, *args, **kwargs):
                pass

        EntityFieldsForm.uicfg_aff.tag_attribute(('CWUser', 'firstname'), MyField)
        try:
            with self.admin_access.web_request() as req:
                form = req.vreg['forms'].select('base', req, entity=req.user)
                self.assertIsInstance(form.field_by_name('firstname', 'subject', req.user.e_schema),
                                      MyField)
        finally:
            EntityFieldsForm.uicfg_aff.del_rtag('CWUser', 'firstname', '*', 'subject')


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""abstract component class and base components definition for CubicWeb web
client
"""


from cubicweb import _

from warnings import warn

from six import PY3, add_metaclass, text_type

from logilab.common.deprecation import class_deprecated, class_renamed, deprecated
from logilab.mtconverter import xml_escape

from cubicweb import Unauthorized, role, target, tags
from cubicweb.schema import display_name
from cubicweb.uilib import js, domid
from cubicweb.utils import json_dumps, js_href
from cubicweb.view import ReloadableMixIn, Component
from cubicweb.predicates import (no_cnx, paginated_rset, one_line_rset,
                                non_final_entity, partial_relation_possible,
                                partial_has_related_entities)
from cubicweb.appobject import AppObject
from cubicweb.web import INTERNAL_FIELD_VALUE, stdmsgs


# abstract base class for navigation components ################################

class NavigationComponent(Component):
    """abstract base class for navigation components"""
    __regid__ = 'navigation'
    __select__ = paginated_rset()

    cw_property_defs = {
        _('visible'):  dict(type='Boolean', default=True,
                            help=_('display the component or not')),
        }

    page_size_property = 'navigation.page-size'
    start_param = '__start'
    stop_param = '__stop'
    page_link_templ = u'<span class="slice"><a href="%s" title="%s">%s</a></span>'
    selected_page_link_templ = u'<span class="selectedSlice"><a href="%s" title="%s">%s</a></span>'
    previous_page_link_templ = next_page_link_templ = page_link_templ

    def __init__(self, req, rset, **kwargs):
        super(NavigationComponent, self).__init__(req, rset=rset, **kwargs)
        self.starting_from = 0
        self.total = rset.rowcount

    def get_page_size(self):
        try:
            return self._page_size
        except AttributeError:
            page_size = self.cw_extra_kwargs.get('page_size')
            if page_size is None:
                try:
                    page_size = int(self._cw.form.get('page_size'))
                except (ValueError, TypeError):
                    # no or invalid value, fall back
                    pass
            if page_size is None:
                page_size = self._cw.property_value(self.page_size_property)
            self._page_size = page_size
            return page_size

    def set_page_size(self, page_size):
        self._page_size = page_size

    page_size = property(get_page_size, set_page_size)

    def page_boundaries(self):
        try:
            stop = int(self._cw.form[self.stop_param]) + 1
            start = int(self._cw.form[self.start_param])
        except KeyError:
            start, stop = 0, self.page_size
        if start >= len(self.cw_rset):
            start, stop = 0, self.page_size
        self.starting_from = start
        return start, stop

    def clean_params(self, params):
        if self.start_param in params:
            del params[self.start_param]
        if self.stop_param in params:
            del params[self.stop_param]

    def page_url(self, path, params, start=None, stop=None):
        params = dict(params)
        params['__fromnavigation'] = 1
        if start is not None:
            params[self.start_param] = start
        if stop is not None:
            params[self.stop_param] = stop
        view = self.cw_extra_kwargs.get('view')
        if view is not None and hasattr(view, 'page_navigation_url'):
            url = view.page_navigation_url(self, path, params)
        elif path in ('json', 'ajax'):
            # 'ajax' is the new correct controller, but the old 'json'
            # controller should still be supported
            url = self.ajax_page_url(**params)
        else:
            url = self._cw.build_url(path, **params)
        # XXX hack to avoid opening a new page containing the evaluation of the
        # js expression on ajax call
        if url.startswith('javascript:'):
            url += '; $.noop();'
        return url

    def ajax_page_url(self, **params):
        divid = params.setdefault('divid', 'pageContent')
        params['rql'] = self.cw_rset.printable_rql()
        return js_href("$(%s).loadxhtml(AJAX_PREFIX_URL, %s, 'get', 'swap')" % (
            json_dumps('#'+divid), js.ajaxFuncArgs('view', params)))

    def page_link(self, path, params, start, stop, content):
        url = xml_escape(self.page_url(path, params, start, stop))
        if start == self.starting_from:
            return self.selected_page_link_templ % (url, content, content)
        return self.page_link_templ % (url, content, content)

    @property
    def prev_icon_url(self):
        return xml_escape(self._cw.data_url('go_prev.png'))

    @property
    def next_icon_url(self):
        return xml_escape(self._cw.data_url('go_next.png'))

    @property
    def no_previous_page_link(self):
        return (u'<img src="%s" alt="%s" class="prevnext_nogo"/>' %
                (self.prev_icon_url, self._cw._('there is no previous page')))

    @property
    def no_next_page_link(self):
        return (u'<img src="%s" alt="%s" class="prevnext_nogo"/>' %
                (self.next_icon_url, self._cw._('there is no next page')))

    @property
    def no_content_prev_link(self):
        return (u'<img src="%s" alt="%s" class="prevnext"/>' % (
                (self.prev_icon_url, self._cw._('no content prev link'))))

    @property
    def no_content_next_link(self):
        return (u'<img src="%s" alt="%s" class="prevnext"/>' %
                (self.next_icon_url, self._cw._('no content next link')))

    def previous_link(self, path, params, content=None, title=_('previous_results')):
        if not content:
            content = self.no_content_prev_link
        start = self.starting_from
        if not start :
            return self.no_previous_page_link
        start = max(0, start - self.page_size)
        stop = start + self.page_size - 1
        url = xml_escape(self.page_url(path, params, start, stop))
        return self.previous_page_link_templ % (url, self._cw._(title), content)

    def next_link(self, path, params, content=None, title=_('next_results')):
        if not content:
            content = self.no_content_next_link
        start = self.starting_from + self.page_size
        if start >= self.total:
            return self.no_next_page_link
        stop = start + self.page_size - 1
        url = xml_escape(self.page_url(path, params, start, stop))
        return self.next_page_link_templ % (url, self._cw._(title), content)

    def render_link_back_to_pagination(self, w):
        """allow to come back to the paginated view"""
        req = self._cw
        params = dict(req.form)
        basepath = req.relative_path(includeparams=False)
        del params['__force_display']
        url = self.page_url(basepath, params)
        w(u'<div class="displayAllLink"><a href="%s">%s</a></div>\n'
          % (xml_escape(url),
             req._('back to pagination (%s results)') % self.page_size))

    def render_link_display_all(self, w):
        """make a link to see them all"""
        req = self._cw
        params = dict(req.form)
        self.clean_params(params)
        basepath = req.relative_path(includeparams=False)
        params['__force_display'] = 1
        params['__fromnavigation'] = 1
        url = self.page_url(basepath, params)
        w(u'<div class="displayAllLink"><a href="%s">%s</a></div>\n'
          % (xml_escape(url), req._('show %s results') % len(self.cw_rset)))

# new contextual components system #############################################

def override_ctx(cls, **kwargs):
    cwpdefs = cls.cw_property_defs.copy()
    cwpdefs['context']  = cwpdefs['context'].copy()
    cwpdefs['context'].update(kwargs)
    return cwpdefs


class EmptyComponent(Exception):
    """some selectable component has actually no content and should not be
    rendered
    """


class Link(object):
    """a link to a view or action in the ui.

    Use this rather than `cw.web.htmlwidgets.BoxLink`.

    Note this class could probably be avoided with a proper DOM on the server
    side.
    """
    newstyle = True

    def __init__(self, href, label, **attrs):
        self.href = href
        self.label = label
        self.attrs = attrs

    def __unicode__(self):
        return tags.a(self.label, href=self.href, **self.attrs)

    if PY3:
        __str__ = __unicode__

    def render(self, w):
        w(tags.a(self.label, href=self.href, **self.attrs))

    def __repr__(self):
        return '<%s: href=%r label=%r %r>' % (self.__class__.__name__,
                                              self.href, self.label, self.attrs)


class Separator(object):
    """a menu separator.

    Use this rather than `cw.web.htmlwidgets.BoxSeparator`.
    """
    newstyle = True

    def render(self, w):
        w(u'<hr class="boxSeparator"/>')


def _bwcompatible_render_item(w, item):
    if hasattr(item, 'render'):
        if getattr(item, 'newstyle', False):
            if isinstance(item, Separator):
                w(u'</ul>')
                item.render(w)
                w(u'<ul>')
            else:
                w(u'<li>')
                item.render(w)
                w(u'</li>')
        else:
            item.render(w) # XXX displays <li> by itself
    else:
        w(u'<li>%s</li>' % item)


class Layout(Component):
    __regid__ = 'component_layout'
    __abstract__ = True

    def init_rendering(self):
        """init view for rendering. Return true if we should go on, false
        if we should stop now.
        """
        view = self.cw_extra_kwargs['view']
        try:
            view.init_rendering()
        except Unauthorized as ex:
            self.warning("can't render %s: %s", view, ex)
            return False
        except EmptyComponent:
            return False
        return True


class LayoutableMixIn(object):
    layout_id = None # to be defined in concret class
    layout_args = {}

    def layout_render(self, w, **kwargs):
        getlayout = self._cw.vreg['components'].select
        layout = getlayout(self.layout_id, self._cw, **self.layout_select_args())
        layout.render(w)

    def layout_select_args(self):
        args  = dict(rset=self.cw_rset, row=self.cw_row, col=self.cw_col,
                     view=self)
        args.update(self.layout_args)
        return args


class CtxComponent(LayoutableMixIn, AppObject):
    """base class for contextual components. The following contexts are
    predefined:

    * boxes: 'left', 'incontext', 'right'
    * section: 'navcontenttop', 'navcontentbottom', 'navtop', 'navbottom'
    * other: 'ctxtoolbar'

    The 'incontext', 'navcontenttop', 'navcontentbottom' and 'ctxtoolbar'
    contexts are handled by the default primary view, others by the default main
    template.

    All subclasses may not support all those contexts (for instance if it can't
    be displayed as box, or as a toolbar icon). You may restrict allowed context
    as follows:

    .. sourcecode:: python

      class MyComponent(CtxComponent):
          cw_property_defs = override_ctx(CtxComponent,
                                          vocabulary=[list of contexts])
          context = 'my default context'

    You can configure a component's default context by simply giving an
    appropriate value to the `context` class attribute, as seen above.
    """
    __registry__ = 'ctxcomponents'
    __select__ = ~no_cnx()

    categories_in_order = ()
    cw_property_defs = {
        _('visible'): dict(type='Boolean', default=True,
                           help=_('display the box or not')),
        _('order'):   dict(type='Int', default=99,
                           help=_('display order of the box')),
        _('context'): dict(type='String', default='left',
                           vocabulary=(_('left'), _('incontext'), _('right'),
                                       _('navtop'), _('navbottom'),
                                       _('navcontenttop'), _('navcontentbottom'),
                                       _('ctxtoolbar')),
                           help=_('context where this component should be displayed')),
        }
    visible = True
    order = 0
    context = 'left'
    contextual = False
    title = None
    layout_id = 'component_layout'

    def render(self, w, **kwargs):
        self.layout_render(w, **kwargs)

    def layout_select_args(self):
        args = super(CtxComponent, self).layout_select_args()
        try:
            # XXX ensure context is given when the component is reloaded through
            # ajax
            args['context'] = self.cw_extra_kwargs['context']
        except KeyError:
            args['context'] = self.cw_propval('context')
        return args

    def init_rendering(self):
        """init rendering callback: that's the good time to check your component
        has some content to display. If not, you can still raise
        :exc:`EmptyComponent` to inform it should be skipped.

        Also, :exc:`Unauthorized` will be caught, logged, then the component
        will be skipped.
        """
        self.items = []

    @property
    def domid(self):
        """return the HTML DOM identifier for this component"""
        return domid(self.__regid__)

    @property
    def cssclass(self):
        """return the CSS class name for this component"""
        return domid(self.__regid__)

    def render_title(self, w):
        """return the title for this component"""
        if self.title:
            w(self._cw._(self.title))

    def render_body(self, w):
        """return the body (content) for this component"""
        raise NotImplementedError()

    def render_items(self, w, items=None, klass=u'boxListing'):
        if items is None:
            items = self.items
        assert items
        w(u'<ul class="%s">' % klass)
        for item in items:
            _bwcompatible_render_item(w, item)
        w(u'</ul>')

    def append(self, item):
        self.items.append(item)

    def action_link(self, action):
        return self.link(self._cw._(action.title), action.url())

    def link(self, title, url, **kwargs):
        if self._cw.selected(url):
            try:
                kwargs['klass'] += ' selected'
            except KeyError:
                kwargs['klass'] = 'selected'
        return Link(url, title, **kwargs)

    def separator(self):
        return Separator()


class EntityCtxComponent(CtxComponent):
    """base class for boxes related to a single entity"""
    __select__ = CtxComponent.__select__ & non_final_entity() & one_line_rset()
    context = 'incontext'
    contextual = True

    def __init__(self, *args, **kwargs):
        super(EntityCtxComponent, self).__init__(*args, **kwargs)
        try:
            entity = kwargs['entity']
        except KeyError:
            entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        self.entity = entity

    def layout_select_args(self):
        args = super(EntityCtxComponent, self).layout_select_args()
        args['entity'] = self.entity
        return args

    @property
    def domid(self):
        return domid(self.__regid__) + text_type(self.entity.eid)

    def lazy_view_holder(self, w, entity, oid, registry='views'):
        """add a holder and return a URL that may be used to replace this
        holder by the html generate by the view specified by registry and
        identifier. Registry defaults to 'views'.
        """
        holderid = '%sHolder' % self.domid
        w(u'<div id="%s"></div>' % holderid)
        params = self.cw_extra_kwargs.copy()
        params.pop('view', None)
        params.pop('entity', None)
        form = params.pop('formparams', {})
        if entity.has_eid():
            eid = entity.eid
        else:
            eid = None
            form['etype'] = entity.cw_etype
            form['tempEid'] = entity.eid
        args = [json_dumps(x) for x in (registry, oid, eid, params)]
        return self._cw.ajax_replace_url(
            holderid, fname='render', arg=args, **form)


# high level abstract classes ##################################################

class RQLCtxComponent(CtxComponent):
    """abstract box for boxes displaying the content of a rql query not related
    to the current result set.

    Notice that this class's init_rendering implemention is overwriting context
    result set (eg `cw_rset`) with the result set returned by execution of
    `to_display_rql()`.
    """
    rql = None

    def to_display_rql(self):
        """return arguments to give to self._cw.execute, as a tuple, to build
        the result set to be displayed by this box.
        """
        assert self.rql is not None, self.__regid__
        return (self.rql,)

    def init_rendering(self):
        super(RQLCtxComponent, self).init_rendering()
        self.cw_rset = self._cw.execute(*self.to_display_rql())
        if not self.cw_rset:
            raise EmptyComponent()

    def render_body(self, w):
        rset = self.cw_rset
        if len(rset[0]) == 2:
            items = []
            for i, (eid, label) in enumerate(rset):
                entity = rset.get_entity(i, 0)
                items.append(self.link(label, entity.absolute_url()))
        else:
            items = [self.link(e.dc_title(), e.absolute_url())
                     for e in rset.entities()]
        self.render_items(w, items)


class EditRelationMixIn(ReloadableMixIn):

    def box_item(self, entity, etarget, fname, label):
        """builds HTML link to edit relation between `entity` and `etarget`"""
        args = {role(self) : entity.eid, target(self): etarget.eid}
        # for each target, provide a link to edit the relation
        jscall = js.cw.utils.callAjaxFuncThenReload(fname,
                                                    self.rtype,
                                                    args['subject'],
                                                    args['object'])
        return u'[<a href="javascript: %s" class="action">%s</a>] %s' % (
            xml_escape(text_type(jscall)), label, etarget.view('incontext'))

    def related_boxitems(self, entity):
        return [self.box_item(entity, etarget, 'delete_relation', u'-')
                for etarget in self.related_entities(entity)]

    def related_entities(self, entity):
        return entity.related(self.rtype, role(self), entities=True)

    def unrelated_boxitems(self, entity):
        return [self.box_item(entity, etarget, 'add_relation', u'+')
                for etarget in self.unrelated_entities(entity)]

    def unrelated_entities(self, entity):
        """returns the list of unrelated entities, using the entity's
        appropriate vocabulary function
        """
        skip = set(text_type(e.eid) for e in entity.related(self.rtype, role(self),
                                                          entities=True))
        skip.add(None)
        skip.add(INTERNAL_FIELD_VALUE)
        filteretype = getattr(self, 'etype', None)
        entities = []
        form = self._cw.vreg['forms'].select('edition', self._cw,
                                             rset=self.cw_rset,
                                             row=self.cw_row or 0)
        field = form.field_by_name(self.rtype, role(self), entity.e_schema)
        for _, eid in field.vocabulary(form):
            if eid not in skip:
                entity = self._cw.entity_from_eid(eid)
                if filteretype is None or entity.cw_etype == filteretype:
                    entities.append(entity)
        return entities

# XXX should be a view usable using uicfg
class EditRelationCtxComponent(EditRelationMixIn, EntityCtxComponent):
    """base class for boxes which let add or remove entities linked by a given
    relation

    subclasses should define at least id, rtype and target class attributes.
    """
    # to be defined in concrete classes
    rtype = None

    def render_title(self, w):
        w(display_name(self._cw, self.rtype, role(self),
                       context=self.entity.cw_etype))

    def render_body(self, w):
        self._cw.add_js('cubicweb.ajax.js')
        related = self.related_boxitems(self.entity)
        unrelated = self.unrelated_boxitems(self.entity)
        self.items.extend(related)
        if related and unrelated:
            self.items.append(u'<hr class="boxSeparator"/>')
        self.items.extend(unrelated)
        self.render_items(w)


class AjaxEditRelationCtxComponent(EntityCtxComponent):
    __select__ = EntityCtxComponent.__select__ & (
        partial_relation_possible(action='add') | partial_has_related_entities())

    # view used to display related entties
    item_vid = 'incontext'
    # values separator when multiple values are allowed
    separator = ','
    # msgid of the message to display when some new relation has been added/removed
    added_msg = None
    removed_msg = None

    # to be defined in concrete classes
    rtype = role = target_etype = None
    # class attributes below *must* be set in concrete classes (additionally to
    # rtype / role [/ target_etype]. They should correspond to js_* methods on
    # the json controller

    # function(eid)
    # -> expected to return a list of values to display as input selector
    #    vocabulary
    fname_vocabulary = None

    # function(eid, value)
    # -> handle the selector's input (eg create necessary entities and/or
    # relations). If the relation is multiple, you'll get a list of value, else
    # a single string value.
    fname_validate = None

    # function(eid, linked entity eid)
    # -> remove the relation
    fname_remove = None

    def __init__(self, *args, **kwargs):
        super(AjaxEditRelationCtxComponent, self).__init__(*args, **kwargs)
        self.rdef = self.entity.e_schema.rdef(self.rtype, self.role, self.target_etype)

    def render_title(self, w):
        w(self.rdef.rtype.display_name(self._cw, self.role,
                                       context=self.entity.cw_etype))

    def add_js_css(self):
        self._cw.add_js(('jquery.ui.js', 'cubicweb.widgets.js'))
        self._cw.add_js(('cubicweb.ajax.js', 'cubicweb.ajax.box.js'))
        self._cw.add_css('jquery.ui.css')
        return True

    def render_body(self, w):
        req = self._cw
        entity = self.entity
        related = entity.related(self.rtype, self.role)
        if self.role == 'subject':
            mayadd = self.rdef.has_perm(req, 'add', fromeid=entity.eid)
        else:
            mayadd = self.rdef.has_perm(req, 'add', toeid=entity.eid)
        js_css_added = False
        if mayadd:
            js_css_added = self.add_js_css()
        _ = req._
        if related:
            maydel = None
            w(u'<table class="ajaxEditRelationTable">')
            for rentity in related.entities():
                if maydel is None:
                    # Only check permission for the first related.
                    if self.role == 'subject':
                        fromeid, toeid = entity.eid, rentity.eid
                    else:
                        fromeid, toeid = rentity.eid, entity.eid
                    maydel = self.rdef.has_perm(
                            req, 'delete', fromeid=fromeid, toeid=toeid)
                # for each related entity, provide a link to remove the relation
                subview = rentity.view(self.item_vid)
                if maydel:
                    if not js_css_added:
                        js_css_added = self.add_js_css()
                    jscall = text_type(js.ajaxBoxRemoveLinkedEntity(
                        self.__regid__, entity.eid, rentity.eid,
                        self.fname_remove,
                        self.removed_msg and _(self.removed_msg)))
                    w(u'<tr><td class="dellink">[<a href="javascript: %s">-</a>]</td>'
                      '<td class="entity"> %s</td></tr>' % (xml_escape(jscall),
                                                            subview))
                else:
                    w(u'<tr><td class="entity">%s</td></tr>' % (subview))
            w(u'</table>')
        else:
            w(_('no related entity'))
        if mayadd:
            multiple = self.rdef.role_cardinality(self.role) in '*+'
            w(u'<table><tr><td>')
            jscall = text_type(js.ajaxBoxShowSelector(
                self.__regid__, entity.eid, self.fname_vocabulary,
                self.fname_validate, self.added_msg and _(self.added_msg),
                _(stdmsgs.BUTTON_OK[0]), _(stdmsgs.BUTTON_CANCEL[0]),
                multiple and self.separator))
            w('<a class="button sglink" href="javascript: %s">%s</a>' % (
                xml_escape(jscall),
                multiple and _('add_relation') or _('update_relation')))
            w(u'</td><td>')
            w(u'<div id="%sHolder"></div>' % self.domid)
            w(u'</td></tr></table>')


class RelatedObjectsCtxComponent(EntityCtxComponent):
    """a contextual component to display entities related to another"""
    __select__ = EntityCtxComponent.__select__ & partial_has_related_entities()
    context = 'navcontentbottom'
    rtype = None
    role = 'subject'

    vid = 'list'

    def render_body(self, w):
        rset = self.entity.related(self.rtype, role(self))
        self._cw.view(self.vid, rset, w=w)


# old contextual components, deprecated ########################################

@add_metaclass(class_deprecated)
class EntityVComponent(Component):
    """abstract base class for additinal components displayed in content
    headers and footer according to:

    * the displayed entity's type
    * a context (currently 'header' or 'footer')

    it should be configured using .accepts, .etype, .rtype, .target and
    .context class attributes
    """
    __deprecation_warning__ = '[3.10] *VComponent classes are deprecated, use *CtxComponent instead (%(cls)s)'

    __registry__ = 'ctxcomponents'
    __select__ = one_line_rset()

    cw_property_defs = {
        _('visible'):  dict(type='Boolean', default=True,
                            help=_('display the component or not')),
        _('order'):    dict(type='Int', default=99,
                            help=_('display order of the component')),
        _('context'):  dict(type='String', default='navtop',
                            vocabulary=(_('navtop'), _('navbottom'),
                                        _('navcontenttop'), _('navcontentbottom'),
                                        _('ctxtoolbar')),
                            help=_('context where this component should be displayed')),
    }

    context = 'navcontentbottom'

    def call(self, view=None):
        if self.cw_rset is None:
            self.entity_call(self.cw_extra_kwargs.pop('entity'))
        else:
            self.cell_call(0, 0, view=view)

    def cell_call(self, row, col, view=None):
        self.entity_call(self.cw_rset.get_entity(row, col), view=view)

    def entity_call(self, entity, view=None):
        raise NotImplementedError()

class RelatedObjectsVComponent(EntityVComponent):
    """a section to display some related entities"""
    __select__ = EntityVComponent.__select__ & partial_has_related_entities()

    vid = 'list'
    # to be defined in concrete classes
    rtype = title = None

    def rql(self):
        """override this method if you want to use a custom rql query"""
        return None

    def cell_call(self, row, col, view=None):
        rql = self.rql()
        if rql is None:
            entity = self.cw_rset.get_entity(row, col)
            rset = entity.related(self.rtype, role(self))
        else:
            eid = self.cw_rset[row][col]
            rset = self._cw.execute(self.rql(), {'x': eid})
        if not rset.rowcount:
            return
        self.w(u'<div class="%s">' % self.cssclass)
        self.w(u'<h4>%s</h4>\n' % self._cw._(self.title).capitalize())
        self.wview(self.vid, rset)
        self.w(u'</div>')
